# -*- coding: utf-8 -*-

import unittest
import boto

TABLE_NAME1 = 'Table-1'
TABLE_NAME2 = 'Table-2'
TABLE_NAME3 = 'Table-3'
TABLE_NAME4 = 'Table-4'
TABLE_NAME_INVALID1 = 'Table-invalid 1'

TABLE_RT = 45
TABLE_WT = 123

TABLE_SCHEMA1 = {
    'hash_key_name': 'hash_key',
    'hash_key_proto_value': int,
    'range_key_name': 'range_key',
    'range_key_proto_value': unicode,
}

TABLE_SCHEMA2 = {
    'hash_key_name': 'hash_key',
    'hash_key_proto_value': int,
}

TABLE_SCHEMA_INVALID1 = {
    'hash_key_name': 'hash_key',
    'hash_key_proto_value': unicode,
}

# Not much can be tested here as most bugs are caught by Boto :)

class TestCreateTable(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_create_table_hash_range(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        table = db.create_table(
            name=TABLE_NAME1,
            schema=db.create_schema(**TABLE_SCHEMA1),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

        self.assertEqual(TABLE_NAME1, table.name)
        self.assertEqual(TABLE_RT, table.read_units)
        self.assertEqual(TABLE_WT, table.write_units)
        self.assertEqual(u'CREATING', table.status)
        self.assertEqual(TABLE_SCHEMA1['hash_key_name'], table.schema.hash_key_name)
        self.assertEqual(TABLE_SCHEMA1['range_key_name'], table.schema.range_key_name)
        self.assertEqual(u'N', table.schema.hash_key_type)
        self.assertEqual(u'S', table.schema.range_key_type)

        data = dynamodb.data
        assert TABLE_NAME1 in data
        table = data[TABLE_NAME1]

        self.assertEqual(TABLE_NAME1, table.name)
        self.assertEqual(TABLE_RT, table.rt)
        self.assertEqual(TABLE_WT, table.wt)
        self.assertEqual(TABLE_SCHEMA1['hash_key_name'], table.hash_key.name)
        self.assertEqual(TABLE_SCHEMA1['range_key_name'], table.range_key.name)
        self.assertEqual(u'N', table.hash_key.typename)
        self.assertEqual(u'S', table.range_key.typename)

    def test_create_table_hash(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        table = db.create_table(
            name=TABLE_NAME2,
            schema=db.create_schema(**TABLE_SCHEMA2),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

        self.assertEqual(TABLE_NAME2, table.name)
        self.assertEqual(TABLE_RT, table.read_units)
        self.assertEqual(TABLE_WT, table.write_units)
        self.assertEqual(u'CREATING', table.status)
        self.assertEqual(TABLE_SCHEMA2['hash_key_name'], table.schema.hash_key_name)
        self.assertEqual(u'N', table.schema.hash_key_type)
        self.assertIsNone(table.schema.range_key_name)
        self.assertIsNone(table.schema.range_key_type)

        data = dynamodb.data
        assert TABLE_NAME2 in data
        table = data[TABLE_NAME2]

        self.assertEqual(TABLE_NAME2, table.name)
        self.assertEqual(TABLE_RT, table.rt)
        self.assertEqual(TABLE_WT, table.wt)
        self.assertEqual(TABLE_SCHEMA2['hash_key_name'], table.hash_key.name)
        self.assertEqual(u'N', table.hash_key.typename)
        self.assertIsNone(table.range_key)

    def test_create_table_twice_fails(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.exception import DynamoDBResponseError

        db = connect_boto_patch()

        #1st
        db.create_table(
            name=TABLE_NAME2,
            schema=db.create_schema(**TABLE_SCHEMA2),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

        #2nd
        self.assertRaisesRegexp(DynamoDBResponseError, 'ResourceInUseException',
        db.create_table,
            name=TABLE_NAME2,
            schema=db.create_schema(**TABLE_SCHEMA2),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )


    def test_create_table_invalid_name(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError as DDBValidationErr

        db = connect_boto_patch()

        assert TABLE_NAME_INVALID1 not in dynamodb.data

        self.assertRaises(DDBValidationErr, db.create_table,
            name=TABLE_NAME_INVALID1,
            schema=db.create_schema(**TABLE_SCHEMA_INVALID1),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

    def test_create_table_reach_max(self):
        from ddbmock import connect_boto_patch, config
        from boto.exception import DynamoDBResponseError

        BK = config.MAX_TABLES
        config.MAX_TABLES = 3

        db = connect_boto_patch()

        #1
        db.create_table(
            name=TABLE_NAME1,
            schema=db.create_schema(**TABLE_SCHEMA2),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

        #2
        db.create_table(
            name=TABLE_NAME2,
            schema=db.create_schema(**TABLE_SCHEMA2),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

        #3
        db.create_table(
            name=TABLE_NAME3,
            schema=db.create_schema(**TABLE_SCHEMA2),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

        #4
        self.assertRaisesRegexp(DynamoDBResponseError, 'LimitExceededException',
        db.create_table,
            name=TABLE_NAME4,
            schema=db.create_schema(**TABLE_SCHEMA2),
            read_units=TABLE_RT,
            write_units=TABLE_WT,
        )

        #restore max
        config.MAX_TABLES = BK



# -*- coding: utf-8 -*-

# pyramid entry point: semantic name
from __future__ import absolute_import

from ddbmock.router import router
from ddbmock.errors import DDBError
from pyramid.response import Response
import json

# wrap routing logic
def pyramid_router(request):
    # extract action
    target = request.headers.get('x-amz-target')
    action = target.split('.', 2)[1] if target is not None else ""

    post = request.json

    # do the job
    try:
        body = router(action, post)
        status = '200 OK'
    except DDBError as e:
        body = e.to_dict()
        status = '{} {}'.format(e.status, e.status_str)

    # prepare output
    response = Response()
    response.body = json.dumps(body)
    response.status = status
    response.content_type = 'application/x-amz-json-1.0'
    response.headers['x-amzn-RequestId'] = post['request_id']  # added by router

    # done
    return response



# -*- coding: utf-8 -*-

# Index overhead. This is added to each Item size when computing table size
INDEX_OVERHEAD = 100

# count: maximum number of tables
MAX_TABLES = 256

### items constraints ###

# bytes: max hash_key_size
MAX_HK_SIZE = 2048
# bytes: max range key size
MAX_RK_SIZE = 1024
# bytes: max item size, not including the index overhead
MAX_ITEM_SIZE = 64*1024

### throughput constraints ###

# value: minimum throughput value
MIN_TP = 1
# value: maximum throughput value
MAX_TP = 10000
# days: min time between 2 throughtput decrease
MIN_TP_DEC_INTERVAL = 1
# percent: max throughput increase per single operation
MAX_TP_INC_CHANGE = 100

### slowness simulation ###

# boolean: enable timers ?
ENABLE_DELAYS = True
# seconds: simulate table creation delay. It will still be available immediately
DELAY_CREATING = 1000
# seconds: simulate table update delay. It will still be available immediately
DELAY_UPDATING = 1000
# seconds: simulate table deletion delay. It will still be available until time is exhauted
DELAY_DELETING = 1000

### Throughput statistics ###

# seconds: datapoint duration
STAT_TP_SAMPLE = 1
# seconds: aggregation period
STAT_TP_AGGREG = 5*60

### Storage specific settings ###

# Storage engine to use ('memory' or 'sqlite')
STORAGE_ENGINE_NAME = 'memory'
# SQLite database location
STORAGE_SQLITE_FILE = 'dynamo.db'



# -*- coding: utf-8 -*-

import unittest, mock
import sqlite3, cPickle as pickle
from ddbmock import config

config.STORAGE_SQLITE_FILE = ':memory:'
TABLE_NAME = 'test_table'

ITEM1 = {"key":"value 1"}
ITEM2 = {"key":"value 2"}
ITEM3 = {"key":"value 3"}
ITEM4 = {"key":"value 4"}

ITEM5 = {"key":"value 5"}
ITEM6 = {"key":"value 6"}

class TestSQLiteStore(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.storage.sqlite import conn

        conn.execute('DROP TABLE IF EXISTS `test_table`')
        conn.execute('''CREATE TABLE `test_table` (
          `hash_key` blob NOT NULL,
          `range_key` blob NOT NULL,
          `data` blob NOT NULL,
          PRIMARY KEY (`hash_key`,`range_key`)
        );''')

        conn.executemany('''INSERT INTO `test_table` VALUES (?, ?, ?)''',
                         [
                            (123, 'toto', buffer(pickle.dumps(ITEM1, 2))),
                            (123, 'titi', buffer(pickle.dumps(ITEM2, 2))),
                            (123, 'tata', buffer(pickle.dumps(ITEM3, 2))),
                            (456, 'toto', buffer(pickle.dumps(ITEM4, 2))),
                         ])

        conn.commit()

        self.conn = conn

    def tearDown(self):
        self.conn.execute('DROP TABLE `test_table`')
        self.conn.commit()

    def test_truncate(self):
        from ddbmock.database.storage.sqlite import Store

        store = Store(TABLE_NAME)
        store.truncate()

        count = self.conn.execute('SELECT Count(*) FROM test_table').fetchone()
        self.assertEqual(0, count[0])

    def test_iter(self):
        from ddbmock.database.storage.sqlite import Store

        store = Store(TABLE_NAME)
        self.assertEqual([ITEM1, ITEM2, ITEM3, ITEM4], list(store))

    def test_get_item(self):
        from ddbmock.database.storage.sqlite import Store

        store = Store(TABLE_NAME)
        self.assertEqual(ITEM2, store[(123, 'titi')])
        self.assertEqual(ITEM4, store[(456, 'toto')])

        self.assertEqual({
                            'toto': ITEM1,
                            'titi': ITEM2,
                            'tata': ITEM3,
                         }, store[(123, None)])

        self.assertRaises(KeyError, store.__getitem__, (404, None))
        self.assertRaises(KeyError, store.__getitem__, (132, '404'))

    def test_del_item(self):
        from ddbmock.database.storage.sqlite import Store

        store = Store(TABLE_NAME)
        del store[123, 'toto']
        del store[123, 'titi']
        del store[404, 'titi']
        del store[456, 'toto']

        # it's not real unit test: I use __iter__ to check
        self.assertEqual([ITEM3], list(store))

    def test_set_item(self):
        from ddbmock.database.storage.sqlite import Store

        store = Store(TABLE_NAME)

        store[123, 'titi'] = ITEM5
        store[456, 'titi'] = ITEM6

        # it's not real unit test: I use __iter__ to check
        self.assertEqual([ITEM1, ITEM3, ITEM4, ITEM5, ITEM6], list(store))



# -*- coding: utf-8 -*-

import unittest


class TestItemFieldComparison(unittest.TestCase):
    def test_eq(self):
        from ddbmock.database.comparison import eq

        self.assertTrue(eq({u'S': u'waldo'}, {u'S': u'waldo'}))
        self.assertFalse(eq({u'S': u'waldo'}, {u'S': u'on-time'}))
        self.assertFalse(eq(None, {u'S': u'on-time'}))

    def test_le(self):
        from ddbmock.database.comparison import le

        self.assertTrue(le({u'S': u'waldo'}, {u'S': u'waldo'}))
        self.assertTrue(le({u'S': u'waldo'}, {u'S': u'waldo1'}))
        self.assertFalse(le({u'S': u'waldo'}, {u'S': u'wald'}))

        self.assertTrue(le({u'N': u'42'}, {u'N': u'42'}))
        self.assertTrue(le({u'N': u'42'}, {u'N': u'42.001'}))
        self.assertFalse(le({u'N': u'42'}, {u'N': u'7'}))

        self.assertFalse(le(None, {u'N': u'7'}))

    def test_lt(self):
        from ddbmock.database.comparison import lt

        self.assertFalse(lt({u'S': u'waldo'}, {u'S': u'waldo'}))
        self.assertTrue(lt({u'S': u'waldo'}, {u'S': u'waldo1'}))
        self.assertFalse(lt({u'S': u'waldo'}, {u'S': u'wald'}))

        self.assertFalse(lt({u'N': u'42'}, {u'N': u'42'}))
        self.assertTrue(lt({u'N': u'42'}, {u'N': u'42.001'}))
        self.assertFalse(lt({u'N': u'42'}, {u'N': u'7'}))

        self.assertFalse(lt(None, {u'N': u'7'}))

    def test_ge(self):
        from ddbmock.database.comparison import ge

        self.assertTrue(ge({u'S': u'waldo'}, {u'S': u'waldo'}))
        self.assertFalse(ge({u'S': u'waldo'}, {u'S': u'waldo1'}))
        self.assertTrue(ge({u'S': u'waldo'}, {u'S': u'wald'}))

        self.assertTrue(ge({u'N': u'42'}, {u'N': u'42'}))
        self.assertFalse(ge({u'N': u'42'}, {u'N': u'42.001'}))
        self.assertTrue(ge({u'N': u'42'}, {u'N': u'7'}))

        self.assertFalse(ge(None, {u'N': u'7'}))

    def test_gt(self):
        from ddbmock.database.comparison import gt

        self.assertFalse(gt({u'S': u'waldo'}, {u'S': u'waldo'}))
        self.assertFalse(gt({u'S': u'waldo'}, {u'S': u'waldo1'}))
        self.assertTrue(gt({u'S': u'waldo'}, {u'S': u'wald'}))

        self.assertFalse(gt({u'N': u'42'}, {u'N': u'42'}))
        self.assertFalse(gt({u'N': u'42'}, {u'N': u'42.001'}))
        self.assertTrue(gt({u'N': u'42'}, {u'N': u'7'}))

        self.assertFalse(gt(None, {u'N': u'7'}))

    def test_begins_with(self):
        from ddbmock.database.comparison import begins_with

        self.assertTrue(begins_with({u'S': u'waldo'}, {u'S': u'waldo'}))
        self.assertTrue(begins_with({u'S': u'waldo'}, {u'S': u'wal'}))
        self.assertFalse(begins_with({u'S': u'waldo'}, {u'S': u'waldo-mario'}))

        self.assertFalse(begins_with(None, {u'S': u'waldo-mario'}))

    def test_between(self):
        from ddbmock.database.comparison import between

        self.assertTrue(between({u'S': u'waldo'}, {u'S': u'wal-mart'}, {u'S': u'walee'}))
        self.assertFalse(between({u'S': u'waldo'}, {u'S': u'wale'}, {u'S': u'walee'}))
        self.assertTrue(between({u'N': u'42'}, {u'N': u'7'}, {u'N': u'123'}))
        self.assertTrue(between({u'N': u'42'}, {u'N': u'42'}, {u'N': u'42'}))
        self.assertFalse(between({u'N': u'42'}, {u'N': u'43'}, {u'N': u'123'}))

        self.assertFalse(between(None, {u'N': u'43'}, {u'N': u'123'}))

    def test_null(self):
        from ddbmock.database.comparison import null

        self.assertTrue(null(None))
        self.assertFalse(null([]))
        self.assertFalse(null({}))
        self.assertFalse(null(False))
        self.assertFalse(null({u'N': u'42'}))

    def test_not_null(self):
        from ddbmock.database.comparison import not_null

        self.assertFalse(not_null(None))
        self.assertTrue(not_null([]))
        self.assertTrue(not_null({}))
        self.assertTrue(not_null(False))
        self.assertTrue(not_null({u'N': u'42'}))

    def test_contains(self):
        from ddbmock.database.comparison import contains

        self.assertTrue(contains({u'S': u'waldo'}, {u'S': u'al'}))
        self.assertFalse(contains({u'S': u'waldo'}, {u'S': u'alee'}))
        self.assertTrue(contains({u'SS': [u'waldo']}, {u'S': u'waldo'}))
        self.assertFalse(contains({u'SS': [u'waldo']}, {u'S': u'al'}))
        self.assertTrue(contains({u'NS': [u'123']}, {u'N': u'123'}))
        self.assertFalse(contains({u'NS': [u'123']}, {u'N': u'12'}))

    def test_not_contains(self):
        from ddbmock.database.comparison import not_contains

        self.assertFalse(not_contains({u'S': u'waldo'}, {u'S': u'al'}))
        self.assertTrue(not_contains({u'S': u'waldo'}, {u'S': u'alee'}))
        self.assertFalse(not_contains({u'SS': [u'waldo']}, {u'S': u'waldo'}))
        self.assertTrue(not_contains({u'SS': [u'waldo']}, {u'S': u'al'}))
        self.assertFalse(not_contains({u'NS': [u'123']}, {u'N': u'123'}))
        self.assertTrue(not_contains({u'NS': [u'123']}, {u'N': u'12'}))

    def test_in(self):
        import ddbmock.database.comparison as c

        # This module is intended to be used reflexively, so... normal behovior
        test_in = getattr(c, 'in')

        pool = [
            {u'S': u'waldo'},
            {u'N': u'123'},
        ]

        self.assertTrue(test_in({u'S': u'waldo'}, *pool))
        self.assertTrue(test_in({u'N': u'123'}, *pool))
        self.assertFalse(test_in({u'S': u'wal'}, *pool))
        self.assertFalse(test_in({u'N': u'12'}, *pool))

    def test_type_mismatch(self):
        import ddbmock.database.comparison as c
        from ddbmock.database.comparison import (
            between, begins_with, eq, lt, le, gt, ge, contains)
        test_in = getattr(c, 'in')

        target = {u'S': u'waldo'}
        target2 = {u'NS': [u'123']}
        rule = {u'N': u'123'}

        self.assertRaises(TypeError, eq, target, rule)
        self.assertRaises(TypeError, lt, target, rule)
        self.assertRaises(TypeError, le, target, rule)
        self.assertRaises(TypeError, gt, target, rule)
        self.assertRaises(TypeError, ge, target, rule)
        self.assertRaises(TypeError, begins_with, target, rule)
        self.assertRaises(TypeError, between, target, rule, rule)
        self.assertRaises(TypeError, contains, target, rule)
        self.assertRaises(TypeError, contains, rule, rule) # can not use contains on number
        self.assertRaises(TypeError, test_in, target2, rule) # target can only be scalar



# -*- coding: utf-8 -*-

import unittest
import boto

TABLE_NAME = 'Table-1'
TABLE_NAME2 = 'Table-2'
TABLE_NAME_404 = 'Waldo'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 200
TABLE_RT3 = 10
TABLE_WT3 = 2564756456
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

class TestUpdateTables(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        t1 = Table(TABLE_NAME,  TABLE_RT, TABLE_WT, hash_key, range_key, status="ACTIVE")
        t2 = Table(TABLE_NAME2, TABLE_RT, TABLE_WT, hash_key, range_key)

        dynamodb.data[TABLE_NAME]  = t1
        dynamodb.data[TABLE_NAME2] = t2

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_update(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        db.layer1.update_table(TABLE_NAME, {'ReadCapacityUnits': TABLE_RT2,
                                            'WriteCapacityUnits': TABLE_WT2})

        data = dynamodb.data
        assert TABLE_NAME in data
        table = data[TABLE_NAME]

        self.assertEqual(TABLE_NAME, table.name)
        self.assertEqual(TABLE_RT2, table.rt)
        self.assertEqual(TABLE_WT2, table.wt)

    def test_update_CREATING_status(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.exception import DynamoDBResponseError

        db = connect_boto_patch()

        self.assertRaisesRegexp(DynamoDBResponseError, 'ResourceInUseException',
                                db.layer1.update_table,
                                TABLE_NAME2, {'ReadCapacityUnits': TABLE_RT2,
                                              'WriteCapacityUnits': TABLE_WT2})

    def test_update_404(self):
        from ddbmock import connect_boto_patch
        from boto.exception import BotoServerError

        db = connect_boto_patch()

        self.assertRaises(BotoServerError, db.layer1.update_table,
                          TABLE_NAME_404,
                          {'ReadCapacityUnits': TABLE_RT2,
                           'WriteCapacityUnits': TABLE_WT2},
                         )

    def test_update_limite_exceeded_propagates(self):
        from ddbmock import connect_boto_patch
        from boto.exception import BotoServerError

        db = connect_boto_patch()

        self.assertRaises(BotoServerError, db.layer1.update_table,
                          TABLE_NAME,
                          {'ReadCapacityUnits': TABLE_RT3,
                           'WriteCapacityUnits': TABLE_WT3},
                         )




# -*- coding: utf-8 -*-

import unittest

HOST = 'samaritain'
PORT = -12321
ENDPOINT = '{}:{}'.format(HOST, PORT)

class TestDdbmockInit(unittest.TestCase):
    def tearDown(self):
        from ddbmock import clean_boto_patch
        clean_boto_patch()

    def test_connect_boto_patch_patch(self):
        from ddbmock import connect_boto_patch, real_boto, noop
        from ddbmock.router.boto import boto_router

        db = connect_boto_patch()

        self.assertEqual(boto_router, db.layer1.make_request.im_func)
        self.assertEqual(noop, db.layer1.__init__.im_func)
        self.assertNotEqual(real_boto['Layer1.make_request'], boto_router)
        self.assertNotEqual(real_boto['Layer1.__init__'], noop)

        # do it twice to make sure it is safe
        db = connect_boto_patch()

        self.assertEqual(boto_router, db.layer1.make_request.im_func)
        self.assertEqual(noop, db.layer1.__init__.im_func)
        self.assertNotEqual(real_boto['Layer1.make_request'], boto_router)
        self.assertNotEqual(real_boto['Layer1.__init__'], noop)

    def test_connect_boto_patch_network(self):
        from ddbmock import connect_boto_network, clean_boto_patch

        clean_boto_patch()
        db = connect_boto_network(HOST, PORT)

        self.assertEqual('ddbmock', db.layer1.region.name)
        self.assertEqual(ENDPOINT, db.layer1.region.endpoint)
        self.assertEqual(ENDPOINT, db.layer1.host)
        self.assertEqual(PORT, db.layer1.port)
        self.assertFalse(db.layer1.is_secure)



# -*- coding: utf-8 -*-

import unittest, mock, json, time

NOW = time.time()

TABLE_NAME = u'Table-1'
TABLE_RT = 45
TABLE_WT = 123
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HEADERS = {
    'x-amz-target': 'dynamodb_20111205.DeleteTable',
    'content-type': 'application/x-amz-json-1.0',
}

# Goal here is not to test the full API, this is done by the Boto tests
class TestDeleteTable(unittest.TestCase):
    @mock.patch("ddbmock.database.table.time")  # Brrr
    def setUp(self, m_time):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        from ddbmock import main
        app = main({})
        from webtest import TestApp
        self.app = TestApp(app)

        m_time.time.return_value = NOW

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)
        t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key, status='ACTIVE')
        dynamodb.data[TABLE_NAME] = t1

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    def test_delete_table(self):
        from ddbmock.database.db import dynamodb

        request = {"TableName": TABLE_NAME}

        expected = {
            u'TableDescription': {
                u'CreationDateTime': NOW,
                u'KeySchema': {
                    u'HashKeyElement': {
                        u'AttributeName': TABLE_HK_NAME,
                        u'AttributeType': TABLE_HK_TYPE,
                    },
                    u'RangeKeyElement': {
                        u'AttributeName': TABLE_RK_NAME,
                        u'AttributeType': TABLE_RK_TYPE,
                    },
                },
                u'ProvisionedThroughput': {
                    u'ReadCapacityUnits': TABLE_RT,
                    u'WriteCapacityUnits': TABLE_WT,
                },
                u'TableName': TABLE_NAME,
                u'TableStatus': u'DELETING',
            }
        }

        res = self.app.post_json('/', request, HEADERS, status=200)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])




# -*- coding: utf-8 -*-

# This module is not imported unless connect_boto is called thus making 'boto'
# an optional dependancy

# boto entry point: semantic name
from __future__ import absolute_import

import json, time, boto

from ddbmock.router import router
from ddbmock.errors import ValidationException, ConditionalCheckFailedException, DDBError
from boto.exception import DynamoDBResponseError
from boto.dynamodb.exceptions import (DynamoDBValidationError as DDBValidationErr,
                                      DynamoDBConditionalCheckFailedError)

# DDB to Boto exception
def _ddbmock_exception_to_boto_exception(err):
    if isinstance(err, ValidationException):
        return DDBValidationErr(err.status, err.status_str, err.to_dict())
    elif isinstance(err, ConditionalCheckFailedException):
        return DynamoDBConditionalCheckFailedError(err.status, err.status_str, err.to_dict())
    else:
        return DynamoDBResponseError(err.status, err.status_str, err.to_dict())


# Boto lib version entry point
def boto_router(self, action, body='', object_hook=None):
    """Replace boto.layer1.make_request to avoid the network layer. It also
    skips the authentication part. The external behavior is strictly compatible"""
    target = '%s_%s.%s' % (self.ServiceName, self.Version, action)
    start = time.time()

    post = json.loads(body)

    try:
        ret = router(action, post)
    except DDBError as e:
        raise _ddbmock_exception_to_boto_exception(e)
    finally:
        boto.log.debug('RequestId: %s', post['request_id'])
        elapsed = (time.time() - start) * 1000
        boto.perflog.info('dynamodb %s: id=%s time=%sms', target, post['request_id'], int(elapsed))

    # FIXME: dump followed by load... can be better...
    ret = json.dumps(ret)
    return json.loads(ret, object_hook=object_hook)



# -*- coding: utf-8 -*-

from ddbmock.errors import ConditionalCheckFailedException, ValidationException
from ddbmock import config
from decimal import Decimal
from math import ceil
from . import comparison


def _decode_field(field):
    """
    Read a field's type and value

    :param field: Raw DynamoDB request field of the form ``{'typename':'value'}``

    :return: (typename, value) string tuple
    """
    return field.items()[0]

class ItemSize(int):
    """
    Utility class to represent an :py:class:`Item` size as bytes or capacity units
    """
    def __add__(self, value):
        """
        Transparently allow addition of ``ItemSize`` values. This is useful for
        all batch requests as ``Scan``, ``Query``, ``BatchWriteItem`` and
        ``BatchReadItem``

        :param value: foreign int compatible value to add

        :return: new :py:class:`ItemSize` value

        :raises: ``TypeError`` if ``value`` is not int compatible
        """
        return ItemSize(int.__add__(self, value))

    def as_units(self):
        """
        Get item size in terms of capacity units. This does *not* include the
        index overhead. Units can *not* be bellow 1 ie: a ``DeleteItem`` on a non
        existing item is *not* free

        :return: number of capacity unit consummed by any operation on this ``ItemSize``
        """
        return max(1, int(ceil((self) / 1024.0)))

    def with_indexing_overhead(self):
        """
        Take the indexing overhead into account. this is especially usefull
        to compute the table disk size as DynamoDB would but it's not included
        in the capacity unit calculation.

        :return: ``ItemSize`` + :py:const:`ddbmock.config.INDEX_OVERHEAD`
        """
        return self + config.INDEX_OVERHEAD


class Item(dict):
    """
    Internal Item representation. The Item is stored in its raw DynamoDB request
    form and no parsing is involved unless specifically needed.

    It adds a couple of handful helpers to the dict class such as DynamoDB actions,
    condition validations and specific size computation.
    """
    def __init__(self, dico={}):
        """
        Load a raw DynamoDb Item and enhance it with ou helpers. Also set the
        cached :py:class:`ItemSize` to ``None`` to mark it as not computed. This
        avoids unnecessary computations on temporary Items.

        :param dico: Raw DynamoDB request Item
        """
        self.update(dico)
        self.size = None

    def filter(self, fields):
        """
        Return a dict containing only the keys specified in ``fields``. If
        ``fields`` evaluates to False (None, empty, ...), the original dict is
        returned untouched.

        Internal :py:class:`ItemSize` of the filtered Item is set to original
        Item size as you pay for the data you operated on, not for what was
        actually sent over the wires.

        :param fields: array of name of keys to keep

        :return: filtered ``Item``
        """
        if fields:
            filtered = Item((k, v) for k, v in self.items() if k in fields)
            filtered.size = self.get_size()  # Filtered or not, you pay for actual size
            return filtered
        return self

    def _apply_action(self, fieldname, action):
        """
        Internal function. Applies a single action to a single field.

        :param fieldname: Valid field name
        :param action: Raw DynamoDB request action specification

        :raises: :py:exc:`ddbmock.errors.ValidationException` whenever attempting an illegual action
        """
        # Rewrite this function, it's disgustting code
        if action[u'Action'] == u"PUT":
            self[fieldname] = action[u'Value']

        if action[u'Action'] == u"DELETE": # Starts to be anoying
            if not fieldname in self:
                return  #shortcut
            if u'Value' not in action:
                del self[fieldname] # Nice and easy part
                return

            typename, value = _decode_field(action[u'Value'])
            ftypename, fvalue = _decode_field(self[fieldname])

            if len(ftypename) != 2:
                raise ValidationException(u"Can not DELETE elements from a non set type. Got {}".format(ftypename))
            if ftypename != typename:
                raise ValidationException(u"Expected type {t} for DELETE from type {t}. Got {}".format(typename, t=ftypename))

            # do the dirty work
            data = set(fvalue).difference(value)
            # if data empty => remove the key
            if not data:
                del self[fieldname]
            else:
                self[fieldname] = {ftypename: list(data)}

        if action[u'Action'] == u"ADD":  # Realy anoying to code :s
            #FIXME: not perfect, action should be different if the item was new
            typename, value = _decode_field(action[u'Value'])
            if fieldname in self:
                ftypename, fvalue = _decode_field(self[fieldname])

                if ftypename == u"N":
                    data = Decimal(value) + Decimal(fvalue)
                    self[fieldname][u"N"] = unicode(data)
                elif ftypename in [u"NS", u"SS", u"BS"]:
                    if ftypename != typename:
                        raise ValidationException(u"Expected type {t} for ADD in type {t}. Got {}".format(typename, t=ftypename))
                    data = set(fvalue).union(value)
                    self[fieldname][typename] = list(data)
                else:
                    raise ValidationException(u"Only N, NS, SS and BS types supports ADD operation. Got {}".format(ftypename))
            else:
                if typename not in [u"N", u"NS"]:
                    raise ValidationException(u"When performing ADD operation on new field, only Numbers or Numbers set are allowed. Got {} of type {}".format(value, typename))
                self[fieldname] = action[u'Value']

    def apply_actions(self, actions):
        """
        Apply ``actions`` to the current item. Mostly used by ``UpdateItem``.
        This also resets the cached item size.

        .. warning:: There is a corner case in ``ADD`` action. It will always behave
            as though the item already existed before that is to say, it the target
            field is a non existing set, it will always start a new one with this
            single value in it. In real DynamoDB, if Item was new, it should fail.

        :param action: Raw DynamoDB request actions specification

        :raises: :py:exc:`ddbmock.errors.ValidationException` whenever attempting an illegual action
        """
        map(self._apply_action, actions.keys(), actions.values())
        self.size = None  # reset cache

    def assert_match_expected(self, expected):
        """
        Make sure this Items matches the ``expected`` values. This may be used
        by any signe item write operation such as ``DeleteItem``, ``UpdateItem``
        and ``PutItem``.

        :param expected: Raw DynamoDB request expected values

        :raises: :py:exc:`ddbmock.errors.ConditionalCheckFailedException` if any of the expected values is not valid
        """
        for fieldname, condition in expected.iteritems():
            if u'Exists' in condition and not condition[u'Exists']:
                if fieldname in self:
                    raise ConditionalCheckFailedException(
                        "Field '{}' should not exist".format(fieldname))
                # *IS* executed but coverage bug
                continue  # pragma: no cover
            if fieldname not in self:
                raise ConditionalCheckFailedException(
                    "Field '{}' should exist".format(fieldname))
            if self[fieldname] != condition[u'Value']:
                raise ConditionalCheckFailedException(
                    "Expected field '{}'' = '{}'. Got '{}'".format(
                    fieldname, condition[u'Value'], self[fieldname]))

    def match(self, conditions):
        """
        Check if the current item matches conditions. Return False if a field is not
        found, or does not match. If condition is None, it is considered to match.

        Condition name are assumed to be valid as Onctuous is in charge of input
        validation. Expect crashes otherwise :)

        :param fieldname: Valid field name
        :param condition: Raw DynamoDB request condition of the form ``{"OPERATOR": FIELDDEFINITION}``

        :return: ``True`` on success or ``False`` on first failure
        """
        for name, condition in conditions.iteritems():
            if not self.field_match(name, condition):
                return False

        return True

    def field_match(self, fieldname, condition):
        """
        Check if a field matches a condition. Return False when field not
        found, or do not match. If condition is None, it is considered to match.

        Condition name are assumed to be valid as Onctuous is in charge of input
        validation. Expect crashes otherwise :)

        :param fieldname: Valid field name
        :param condition: Raw DynamoDB request condition of the form ``{"OPERATOR": FIELDDEFINITION}``

        :return: ``True`` on success
        """
        # Arcording to specif, no condition means match
        if condition is None:
            return True

        # read the item
        if fieldname not in self:
            value = None
        else:
            value = self[fieldname]

        # Load the test operator from the comparison module. Thanks to input
        # validation, no try/except required
        condition_operator = condition[u'ComparisonOperator'].lower()
        operator = getattr(comparison, condition_operator)
        return operator(value, *condition[u'AttributeValueList'])

    def read_key(self, key, name=None, max_size=0):
        """
        Provided ``key``, read field value at ``name`` or ``key.name`` if not
        specified.

        :param key: ``Key`` or ``PrimaryKey`` to read
        :param name: override name field of key
        :param max_size: if specified, check that the item is bellow a treshold

        :return: field value at ``key``

        :raises: :py:exc:`ddbmock.errors.ValidationException` if field does not exist, type does not match or is above ``max_size``
        """
        if key is None:
            return False
        if name is None:
            name = key.name

        try:
            field = self[name]
        except KeyError:
            raise ValidationException(u'Field {} not found'.format(name))

        if max_size:
            size = self.get_field_size(name)
            if size > max_size:
                raise ValidationException(u'Field {} is over {} bytes limit. Got {}'.format(name, max_size, size))

        return key.read(field)

    def _internal_item_size(self, base_type, value):
        """
        Internal DynamoDB field size computation. ``base_type`` is assumed to
        be valid as it went through Onctous before and this helper is only
        supposed to be called internally.

        :param base_type: valid base type. Must be in ``['N', 'S', 'B']``.
        :param value: compute the size of this value.
        """
        if base_type == 'N': return 8 # assumes "double" internal type on ddb side
        if base_type == 'S': return len(value.encode('utf-8'))
        if base_type == 'B': return len(value.encode('utf-8'))*3/4 # base64 overhead

    def get_field_size(self, fieldname):
        """
        Compute field size in bytes.

        :param fieldname: Valid field name

        :return: Size of the field in bytes or 0 if the field was not found. Remember that empty fields are represented as missing values in DynamoDB.
        """
        if not fieldname in self:
            return 0

        typename, value = _decode_field(self[fieldname])
        base_type = typename[0]

        if len(typename) == 1:
            value_size = self._internal_item_size(base_type, value)
        else:
            value_size = 0
            for v in value:
                value_size += self._internal_item_size(base_type, v)

        return value_size

    def get_size(self):
        """
        Compute Item size as DynamoDB would. This is especially useful for
        enforcing the 64kb per item limit as well as the capacityUnit cost.

        .. note:: the result is cached for efficiency. If you ever happend to
            directly edit values for any reason, do not forget to invalidate the
            cache: ``self.size=None``

        :return: :py:class:`ItemSize` DynamoDB item size in bytes
        """

        # Check cache and compute
        if self.size is None:
            size = 0

            for key in self.keys():
                size += self._internal_item_size('S', key)
                size += self.get_field_size(key)

            self.size = size

        return ItemSize(self.size)

    def __sub__(self, other):
        """
        Utility function to compute a 'diff' of 2 Items. All fields of ``self``
        (left operand) identical to those of ``other`` (right operand) are dicarded.
        The other fields from ``self`` are kept. This proves to be extremely
        useful to support ``ALL_NEW`` and ``ALL_OLD`` return specification of
        ``UpdateItem`` in a clean and readable manner.

        :param other: ``Item`` to be used as filter

        :return: dict with fields of ``self`` not in or different from ``other``
        """
        # Thanks mnoel :)
        return {k:v for k,v in self.iteritems() if k not in other or v != other[k]}



# -*- coding: utf-8 -*-

# All validations are performed on *incomming* data => already done :)

from ddbmock.errors import ValidationException

class Key(object):
    """
    Abstraction layer over DynamoDB Keys in :py:class:`ddbmock.database.item.Item`
    """
    def __init__(self, name, typename):
        """
        High level Python constructor

        :param name: Valid key name. No further checks are performed.
        :param typename: Valid key typename. No further checks are performed.
        """
        self.name = name
        self.typename = typename

    def read(self, key):
        """
        Parse a key as specified by DynamoDB API and return its value as long as
        its typename matches :py:attr:`typename`

        :param key: Raw DynamoDB request key.

        :return: the value of the key

        :raises: :py:exc:`ddbmock.errors.ValidationException` if field types does not match

        """
        typename, value = key.items()[0]
        if self.typename != typename:
            raise ValidationException('Expected key type = {} for field {}. Got {}'.format(
                self.typename, self.name, typename))

        return value

    def to_dict(self):
        """
        Return the key as a Python dict.

        :return: Serialized version of the key definition metadata compatible with DynamoDB API syntax.
        """
        return {
            "AttributeName": self.name,
            "AttributeType": self.typename,
        }

    @classmethod
    def from_dict(cls, data):
        """
        Alternate constructor which deciphers raw DynamoDB request data before
        ultimately calling regular ``__init__`` method.

        See :py:meth:`__init__` for more insight.

        :param data: raw DynamoDB request data.

        :return: fully initialized :py:class:`Key` instance
        """
        return cls(data[u'AttributeName'], data[u'AttributeType'])

class PrimaryKey(Key):
    """Special marker to provide distinction between regulat Keys and PrimaryKey"""
    pass



# -*- coding: utf-8 -*-

import logging
from ddbmock import config, utils

# Much too slow otherwise
config.DELAY_CREATING = 0.02
config.DELAY_UPDATING = 0.02
config.DELAY_DELETING = 0.02

# Configure logging
utils.tp_logger.addHandler(logging.StreamHandler())



# -*- coding: utf-8 -*-

import unittest
from decimal import Decimal

# tests
# - number validator
# - dynamodb_api_validate errors

ACTION_404 = "!~I bet this route won't ever exist~!"
POST = {"toto":"titi"}

class TestValidators(unittest.TestCase):
    def test_number_max_exp(self):
        from ddbmock.validators.types import Precision
        from onctuous import Invalid

        self.assertEqual(Decimal("382.1E4"), Precision(max_exp=3)(Decimal("382.1E4")))
        self.assertRaisesRegexp(Invalid, "bigger",
                                Precision(max_exp=3),
                                Decimal("395.1E5"))

    def test_number_min_exp(self):
        from ddbmock.validators.types import Precision
        from onctuous import Invalid

        self.assertEqual(Decimal("-50.52"), Precision(min_exp=-2)(Decimal("-50.52")))
        self.assertRaisesRegexp(Invalid, "smaller",
                                Precision(min_exp=-2),
                                Decimal("-50.521"))

    def test_number_max_digits(self):
        from ddbmock.validators.types import Precision
        from onctuous import Invalid

        self.assertEqual(Decimal("-1.21E50"), Precision(max_digits=3)(Decimal("-1.21E50")))
        self.assertRaisesRegexp(Invalid, "digits",
                                Precision(max_digits=3),
                                Decimal("-1.211E50"))

    def test_validator_not_found(self):
        from ddbmock.validators import dynamodb_api_validate

        self.assertEqual(POST, dynamodb_api_validate(ACTION_404, POST))



# -*- coding: utf-8 -*-

import unittest, mock

# tests
# - delete callback
# - create table persist schema

TABLE_NAME = "tabloid"
TABLE_NAME2 = "razoroid"

TABLE_RT = 45
TABLE_WT = 123

HASH_KEY = {"AttributeName":"hash_key","AttributeType":"N"}
RANGE_KEY = {"AttributeName":"range_key","AttributeType":"S"}

class TestDB(unittest.TestCase):
    def setUp(self):
        from ddbmock.database import dynamodb

        self.t1 = mock.Mock()
        self.t1.name = TABLE_NAME

        dynamodb.data[TABLE_NAME] = self.t1
        dynamodb.store[TABLE_NAME, False] = dynamodb.data[TABLE_NAME]

    def test_internal_delete_table(self):
        from ddbmock.database import dynamodb

        # make sure that deleting another table with the same name does nothing
        t2 = mock.Mock()
        t2.name = TABLE_NAME

        # delete a table
        dynamodb._internal_delete_table(t2)
        self.assertIn(TABLE_NAME, dynamodb.data)

        # make sure deleting already deleted table does not harm
        dynamodb._internal_delete_table(self.t1)

    @mock.patch('ddbmock.database.db.schedule_action')
    def test_delete_table(self, m_schedule):
        from ddbmock import config
        from ddbmock.database import dynamodb

        dynamodb.delete_table(TABLE_NAME)
        dynamodb.data[TABLE_NAME].delete.assert_called_with()
        m_schedule.assert_called_with(config.DELAY_DELETING, dynamodb._internal_delete_table, [dynamodb.data[TABLE_NAME]])

    @mock.patch('ddbmock.database.db.dynamodb.store')
    def test_create_table_saves_schema(self, m_store):
        from ddbmock.database.db import dynamodb

        data = {
            "TableName": TABLE_NAME2,
            "KeySchema": {
                "HashKeyElement": HASH_KEY,
                "RangeKeyElement": RANGE_KEY,
            },
            "ProvisionedThroughput": {
                "ReadCapacityUnits": TABLE_RT,
                "WriteCapacityUnits": TABLE_WT,
            }
        }

        dynamodb.create_table(TABLE_NAME2, data)

        m_store.__setitem__.assert_called_with(
                                               (TABLE_NAME2, False),
                                               dynamodb.data[TABLE_NAME2],
                                              )

    @mock.patch('ddbmock.database.db.Store')
    def test_init_reloads_schema(self, m_store):
        from ddbmock.database.db import DynamoDB
        old_internal_state = DynamoDB._shared_data
        DynamoDB._shared_data = {'data': {}, 'store': None}

        table1 = mock.Mock()
        table2 = mock.Mock()
        table1.name = TABLE_NAME
        table2.name = TABLE_NAME2
        m_store.return_value.__iter__.return_value = [table1, table2]

        dynamodb = DynamoDB()

        m_store.assert_called_with('~*schema*~')
        self.assertIn(TABLE_NAME, dynamodb.data)
        self.assertIn(TABLE_NAME2, dynamodb.data)
        self.assertEqual(table1, dynamodb.data[TABLE_NAME])
        self.assertEqual(table2, dynamodb.data[TABLE_NAME2])


        DynamoDB._shared_data = old_internal_state



# -*- coding: utf-8 -*-

import unittest, json

TABLE_NAME1 = 'Table-1'

TABLE_RT = 45
TABLE_WT = 123

TABLE_NAME = 'Table-HR'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE1 = u'123'
HK_VALUE2 = u'456'
HK_VALUE3 = u'789'
RK_VALUE1 = u'Waldo-1'
RK_VALUE2 = u'Waldo-2'
RK_VALUE3 = u'Waldo-3'
RK_VALUE4 = u'Waldo-4'
RK_VALUE5 = u'Waldo-5'


ITEM1 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE1},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE1},
    u'relevant_data': {u'S': u'tata'},
}
ITEM2 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE1},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE2},
    u'relevant_data': {u'S': u'tete'},
}
ITEM3 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE2},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE3},
    u'relevant_data': {u'S': u'titi'},
}
ITEM4 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE3},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE4},
    u'relevant_data': {u'S': u'toto'},
}
ITEM5 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE3},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE5},
    u'relevant_data': {u'S': u'tutu'},
}

HEADERS = {
    'x-amz-target': 'dynamodb_20111205.Scan',
    'content-type': 'application/x-amz-json-1.0',
}

# Goal here is not to test the full API, this is done by the Boto tests
class TestScan(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        from ddbmock.database.db import dynamodb
        from ddbmock import main
        app = main({})
        from webtest import TestApp
        self.app = TestApp(app)

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        self.t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key)

        dynamodb.data[TABLE_NAME]  = self.t1

        self.t1.put(ITEM1, {})
        self.t1.put(ITEM2, {})
        self.t1.put(ITEM3, {})
        self.t1.put(ITEM4, {})
        self.t1.put(ITEM5, {})

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    def test_scan_condition_filter_fields(self):
        from ddbmock.database.db import dynamodb

        request = {
            "TableName": TABLE_NAME,
            "ScanFilter": {
                "relevant_data": {
                    "AttributeValueList": [{"S":"toto"},{"S":"titi"},{"S":"tata"}],
                    "ComparisonOperator": "IN",
                },
            },
            "AttributesToGet": [u'relevant_data'],
        }

        expected = {
            u"Count": 3,
            u"ScannedCount": 5,
            u"Items": [
                {u"relevant_data": {u"S": u"tata"}},
                {u"relevant_data": {u"S": u"toto"}},
                {u"relevant_data": {u"S": u"titi"}},
            ],
            u"ConsumedCapacityUnits": 0.5,
        }

        # Protocol check
        res = self.app.post_json('/', request, HEADERS, status=200)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])

    def test_scan_count_and_attrs_to_get_fails(self):
        from ddbmock.database.db import dynamodb

        request = {
            "TableName": TABLE_NAME,
            "ScanFilter": {
                "relevant_data": {
                    "AttributeValueList": [{"S":"toto"},{"S":"titi"},{"S":"tata"}],
                    "ComparisonOperator": "IN",
                },
            },
            "AttributesToGet": [u'relevant_data'],
            "Count": True,
        }

        expected = {
            u'__type': u'com.amazonaws.dynamodb.v20111205#ValidationException',
            u'message': u'Can not filter fields when only count is requested'
        }

        # Protocol check
        res = self.app.post_json('/', request, HEADERS, status=400)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])



# -*- coding: utf-8 -*-

from . import load_table
from ddbmock.utils import push_write_throughput

@load_table
def delete_item(post, table):
    item = table.delete_item(post[u'Key'], post[u'Expected'])

    capacity = item.get_size().as_units()
    push_write_throughput(table.name, capacity)

    if post[u'ReturnValues'] == "ALL_OLD":
        return {
            "ConsumedCapacityUnits": capacity,
            "Attributes": item,
        }
    else:
        return {
            "ConsumedCapacityUnits": capacity,
        }



# -*- coding: utf-8 -*-

from pyramid.config import Configurator
from ddbmock.router.pyramid import pyramid_router

def noop(*args, **kwargs): pass

# Pyramid entry point
def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    config = Configurator(settings=settings)

    # Insert router as '/' route. This is because all DDB URL are '/' (!)
    config.add_route('pyramid_router', '')
    config.add_view(pyramid_router, route_name='pyramid_router')

    return config.make_wsgi_app()

# Regular "over the network" connection wrapper.
def connect_boto_network(host='localhost', port=6543):
    """Connect to ddbmock launched in *server* mode via boto"""
    import boto
    from boto.regioninfo import RegionInfo
    endpoint = '{}:{}'.format(host,port)
    region = RegionInfo(name='ddbmock', endpoint=endpoint)
    return boto.connect_dynamodb(region=region, port=port, is_secure=False)

# Monkey patch magic, required for the Boto entry point
# Request hijacking Yeah !
real_boto = {}

def connect_boto_patch():
    """Connect to ddbmock as a library via boto"""
    import boto

    if real_boto:
        return boto.connect_dynamodb()

    from boto.dynamodb.layer1 import Layer1
    from router.boto import boto_router

    # Backup real functions for potential cleanup
    real_boto['Layer1.make_request'] = Layer1.make_request
    real_boto['Layer1.__init__'] = Layer1.__init__

    # Bypass network *and* authentication
    Layer1.make_request = boto_router
    Layer1.__init__ = noop

    # Just one more shortcut
    return boto.connect_dynamodb()

def clean_boto_patch():
    """Restore real boto code"""
    if real_boto:
        from boto.dynamodb.layer1 import Layer1

        Layer1.make_request = real_boto['Layer1.make_request']
        Layer1.__init__ = real_boto['Layer1.__init__']

        real_boto.clear()



# -*- coding: utf-8 -*-

from .key import Key, PrimaryKey
from .item import Item, ItemSize
from .storage import Store
from collections import defaultdict, namedtuple
from threading import Lock
from ddbmock import config
from ddbmock.errors import ValidationException, LimitExceededException, ResourceInUseException
from ddbmock.utils import schedule_action
import time, copy, datetime

# All validations are performed on *incomming* data => already done :)

# items: array
# size: ItemSize
Results = namedtuple('Results', ['items', 'size', 'last_key', 'scanned'])

class Table(object):
    """
    Table abstraction. Actual :py:class:`ddbmock.database.item.Item` are stored
    in :py:attr:`store`.
    """
    def __init__(self, name, rt, wt, hash_key, range_key, status='CREATING'):
        """
        Create a new ``Table``. When manually creating a table, make sure you
        registered it in :py:class:`ddbmock.database.db.DynamoDB` with a something
        like ``dynamodb.data[name] = Table(name, "...")``.

        Even though there are :py:const:`DELAY_CREATING` seconds before the status
        is updated to ``ACTIVE``, the table is immediately available. This is a
        slight difference with real DynamooDB to ease unit and functionnal tests.

        :param name: Valid table name. No further checks are performed.
        :param rt: Provisioned read throughput.
        :param wt: Provisioned write throughput.
        :param hash_key: :py:class:`ddbmock.database.key.Key` instance describe the ``hash_key``
        :param hash_key: :py:class:`ddbmock.database.key.Key` instance describe the ``range_key`` or ``None`` if table has no ``range_key``
        :param status: (optional) Valid initial table status. If Table needd to be avaible immediately, use ``ACTIVE``, otherwise, leave default value.

        .. note:: ``rt`` and ``wt`` are only used by ``DescribeTable`` and ``UpdateTable``. No throttling is nor will ever be done.
        """
        self.name = name
        self.rt = rt
        self.wt = wt
        self.hash_key = hash_key
        self.range_key = range_key
        self.status = status

        self.store = Store(name)
        self.write_lock = Lock()

        self.creation_time = time.time()
        self.last_increase_time = 0
        self.last_decrease_time = 0
        self.count = 0

        schedule_action(config.DELAY_CREATING, self.activate)

    def truncate(self):
        """
        Remove all Items from this table. This is like a reset. Might be very
        usefull in unit and functional tests.
        """
        self.store.truncate()

    def delete(self):
        """
        If the table was ``ACTIVE``, update its state to ``DELETING``. This is
        not a destructor, only a sate updater and the Table instance will still
        be valid afterward. In all othercases, raise :py:exc:`ddbmock.errors.ResourceInUseException.`

        If you want to perform the full table delete cycle, please use
        :py:meth:`ddbmock.database.db.DynamoDB.delete_table()` instead

        :raises: :py:exc:`ddbmock.errors.ResourceInUseException` is the table was not in ``Active`` state
        """
        if self.status != "ACTIVE":
            raise ResourceInUseException("Table {} is in {} state. Can not DELETE.".format(self.name, self.status))

        self.status = "DELETING"

    def activate(self):
        """
        Unconditionnaly set Table status to ``ACTIVE``. This method is automatically
        called by the constructor once :py:const:`DELAY_CREATING` is over.
        """
        self.status = "ACTIVE"

    def update_throughput(self, rt, wt):
        """
        Update table throughput. Same conditions and limitations as real DynamoDB
        applies:

        - No more that 1 decrease operation per UTC day.
        - No more than doubling throughput at once.
        - Table must be in ``ACTIVE`` state.

        Table status is then set to ``UPDATING`` until :py:const:`DELAY_UPDATING`
        delay is over. Like real DynamoDB, the Table can still be used during
        this period

        :param rt: New read throughput
        :param wt: New write throughput

        :raises: :py:exc:`ddbmock.errors.ResourceInUseException` if table was not in ``ACTIVE`` state
        :raises: :py:exc:`ddbmock.errors.LimitExceededException` if the other above conditions are not met.

        """
        if self.status != "ACTIVE":
            raise ResourceInUseException("Table {} is in {} state. Can not UPDATE.".format(self.name, self.status))

        # is decrease ?
        if self.rt > rt or self.wt > wt:
            current_time = time.time()
            current_date = datetime.date.fromtimestamp(current_time)
            last_decrease = datetime.date.fromtimestamp(self.last_decrease_time)
            if (current_date - last_decrease).days < config.MIN_TP_DEC_INTERVAL:
                last = datetime.datetime.fromtimestamp(self.last_decrease_time)
                current = datetime.datetime.fromtimestamp(current_time)
                raise LimitExceededException("Subscriber limit exceeded: Provisioned throughput can be decreased only once within the {} day. Last decrease time: Tuesday, {}. Request time: {}".format(config.MIN_TP_DEC_INTERVAL, last, current))
            self.last_decrease_time = current_time

        # is increase ?
        if self.rt < rt or self.wt < wt:
            if (rt - self.rt)/float(self.rt)*100 > config.MAX_TP_INC_CHANGE:
                raise LimitExceededException('Requested provisioned throughput change is not allowed. The ReadCapacityUnits change must be at most {} percent of current value. Current ReadCapacityUnits provisioned for the table: {}. Requested ReadCapacityUnits: {}.'.format(config.MAX_TP_INC_CHANGE, self.rt, rt))
            if (wt - self.wt)/float(self.wt)*100 > config.MAX_TP_INC_CHANGE:
                raise LimitExceededException('Requested provisioned throughput change is not allowed. The WriteCapacityUnits change must be at most {} percent of current value. Current WriteCapacityUnits provisioned for the table: {}. Requested WriteCapacityUnits: {}.'.format(config.MAX_TP_INC_CHANGE, self.wt, wt))
            self.last_increase_time = time.time()

        # real work
        self.status = "UPDATING"

        self.rt = rt
        self.wt = wt

        schedule_action(config.DELAY_UPDATING, self.activate)

    def delete_item(self, key, expected):
        """
        Delete item at ``key`` from the databse provided that it matches ``expected``
        values.

        This operation is atomic and blocks all other pending write operations.

        :param key: Raw DynamoDB request hash and range key dict.
        :param expected: Raw DynamoDB request conditions.

        :return: deepcopy of :py:class:`ddbmock.database.item.Item` as it was before deletion.

        :raises: :py:exc:`ddbmock.errors.ConditionalCheckFailedException` if conditions are not met.
        """
        key = Item(key)
        hash_key = key.read_key(self.hash_key, u'HashKeyElement')
        range_key = key.read_key(self.range_key, u'RangeKeyElement')

        with self.write_lock:
            try:
                old = copy.deepcopy(self.store[hash_key, range_key])
            except KeyError:
                return Item()

            old.assert_match_expected(expected)
            del self.store[hash_key, range_key]

        self.count -= 1
        return old

    def update_item(self, key, actions, expected):
        """
        Apply ``actions`` to item at ``key`` provided that it matches ``expected``.

        This operation is atomic and blocks all other pending write operations.

        :param key: Raw DynamoDB request hash and range key dict.
        :param actions: Raw DynamoDB request actions.
        :param expected: Raw DynamoDB request conditions.

        :return: both deepcopies of :py:class:`ddbmock.database.item.Item` as it was (before, after) the update.

        :raises: :py:exc:`ddbmock.errors.ConditionalCheckFailedException` if conditions are not met.
        :raises: :py:exc:`ddbmock.errors.ValidationException` if ``actions`` attempted to modify the key or the resulting Item is biggere than :py:const:`config.MAX_ITEM_SIZE`
        """
        key = Item(key)
        hash_key = key.read_key(self.hash_key, u'HashKeyElement', max_size=config.MAX_HK_SIZE)
        range_key = key.read_key(self.range_key, u'RangeKeyElement', max_size=config.MAX_RK_SIZE)

        with self.write_lock:
            # Need a deep copy as we will *modify* it
            try:
                old = self.store[hash_key, range_key]
                new = copy.deepcopy(old)
                old.assert_match_expected(expected)
            except KeyError:
                # Item was not in the DB yet
                old = Item()
                new = Item()
                self.count += 1
                # append the keys
                new[self.hash_key.name] = key['HashKeyElement']
                if self.range_key is not None:
                    new[self.range_key.name] = key['RangeKeyElement']


            # Make sure we are not altering a key
            if self.hash_key.name in actions:
                raise ValidationException("UpdateItem can not alter the hash_key.")
            if self.range_key is not None and self.range_key.name in actions:
                raise ValidationException("UpdateItem can not alter the range_key.")

            new.apply_actions(actions)
            self.store[hash_key, range_key] = new

            size = new.get_size()
            if size > config.MAX_ITEM_SIZE:
                self.store[hash_key, range_key] = old  # roll back
                raise ValidationException("Items must be smaller than {} bytes. Got {} after applying update".format(config.MAX_ITEM_SIZE, size))

        return old, new

    def put(self, item, expected):
        """
        Save ``item`` in the database provided that ``expected`` matches. Even
        though DynamoDB ``UpdateItem`` operation only supports returning ``ALL_OLD``
        or ``NONE``, this method returns both ``old`` and ``new`` values as the
        throughput, computed in the view, takes the maximum of both size into
        account.

        This operation is atomic and blocks all other pending write operations.

        :param item: Raw DynamoDB request item.
        :param expected: Raw DynamoDB request conditions.

        :return: both deepcopies of :py:class:`ddbmock.database.item.Item` as it was (before, after) the update or empty item if not found.

        :raises: :py:exc:`ddbmock.errors.ConditionalCheckFailedException` if conditions are not met.
        """
        item = Item(item)

        if item.get_size() > config.MAX_ITEM_SIZE:
            raise ValidationException("Items must be smaller than {} bytes. Got {}".format(config.MAX_ITEM_SIZE, item.get_size()))

        hash_key = item.read_key(self.hash_key, max_size=config.MAX_HK_SIZE)
        range_key = item.read_key(self.range_key, max_size=config.MAX_RK_SIZE)

        with self.write_lock:
            try:
                old = self.store[hash_key, range_key]
                old.assert_match_expected(expected)
            except KeyError:
                # Item was not in the DB yet
                self.count += 1
                old = Item()

            self.store[hash_key, range_key] = item
            new = copy.deepcopy(item)

        return old, new

    def get(self, key, fields):
        """
        Get ``fields`` from :py:class:`ddbmock.database.item.Item` at ``key``.

        :param key: Raw DynamoDB request key.
        :param fields: Raw DynamoDB request array of field names to return. Empty to return all.

        :return: reference to :py:class:`ddbmock.database.item.Item` at ``key`` or ``None`` when not found

        :raises: :py:exc:`ddbmock.errors.ValidationException` if a ``range_key`` was provided while table has none.
        """
        key = Item(key)
        hash_key = key.read_key(self.hash_key, u'HashKeyElement')
        range_key = key.read_key(self.range_key, u'RangeKeyElement')

        if self.range_key is None and u'RangeKeyElement' in key:
            raise ValidationException("Table {} has no range_key".format(self.name))

        try:
            item = self.store[hash_key, range_key]
            return item.filter(fields)
        except KeyError:
            # Item was not in the DB yet
            return None

    def query(self, hash_key, rk_condition, fields, start, reverse, limit):
        """
        Return ``fields`` of all items with provided ``hash_key`` whose ``range_key``
        matches ``rk_condition``.

        :param hash_key: Raw DynamoDB request hash_key.
        :param rk_condition: Raw DynamoDB request ``range_key`` condition.
        :param fields: Raw DynamoDB request array of field names to return. Empty to return all.
        :param start: Raw DynamoDB request key of the first item to scan. Empty array to indicate first item.
        :param reverse: Set to ``True`` to parse the range keys backward.
        :param limit: Maximum number of items to return in this batch. Set to 0 or less for no maximum.

        :return: Results(results, cumulated_size, last_key)

        :raises: :py:exc:`ddbmock.errors.ValidationException` if ``start['HashKeyElement']`` is not ``hash_key``
        """
        #FIXME: naive implementation
        #FIXME: what if an item disappears during the operation ?
        #TODO:
        # - size limit

        hash_value = self.hash_key.read(hash_key)
        rk_name = self.range_key.name
        size = ItemSize(0)
        good_item_count = 0
        results = []
        lek = None

        if start and start['HashKeyElement'] != hash_key:
            raise ValidationException("'HashKeyElement' element of 'ExclusiveStartKey' must be the same as the hash_key. Expected {}, got {}".format(hash_key, start['HashKeyElement']))

        data = self.store[hash_value, None]

        keys = sorted(data.keys())

        if reverse:
            keys.reverse()

        if start:
            first_key = self.range_key.read(start['RangeKeyElement'])
            index = keys.index(first_key) + 1  # May raise ValueError but that's OK
            keys = keys[index:]

        for key in keys:
            item = data[key]

            if item.field_match(rk_name, rk_condition):
                good_item_count += 1
                size += item.get_size()
                results.append(item.filter(fields))

            if good_item_count == limit:
                lek = {
                    u'HashKeyElement': hash_key,
                    u'RangeKeyElement': item[rk_name],
                }
                break

        return Results(results, size, lek, -1)

    def scan(self, scan_conditions, fields, start, limit):
        """
        Return ``fields`` of all items matching ``scan_conditions``. No matter the
        ``start`` key, ``scan`` allways starts from teh beginning so that it might
        be quite slow.

        :param scan_conditions: Raw DynamoDB request conditions.
        :param fields: Raw DynamoDB request array of field names to return. Empty to return all.
        :param start: Raw DynamoDB request key of the first item to scan. Empty array to indicate first item.
        :param limit: Maximum number of items to return in this batch. Set to 0 or less for no maximum.

        :return: Results(results, cumulated_size, last_key, scanned_count)
        """
        #FIXME: naive implementation (too)
        #TODO:
        # - esk
        # - size limit

        size = ItemSize(0)
        scanned = 0
        lek = {}
        results = []
        skip = start is not None
        hk_name = self.hash_key.name
        rk_name = self.range_key.name if self.range_key else None

        for item in self.store:
            # first, skip all items before start
            if skip:
                if start['HashKeyElement'] != item[hk_name]:
                    continue
                if rk_name and start['RangeKeyElement'] != item[rk_name]:
                    continue
                skip = False
                continue

            # match filters ?
            if item.match(scan_conditions):
                results.append(item.filter(fields))

            # update stats
            size += item.get_size()
            scanned += 1

            # quit ?
            if scanned == limit:
                lek[u'HashKeyElement'] = item[hk_name]
                if rk_name:
                    lek[u'RangeKeyElement'] = item[rk_name]
                break

        return Results(results, size, lek, scanned)

    @classmethod
    def from_dict(cls, data):
        """
        Alternate constructor which deciphers raw DynamoDB request data before
        ultimately calling regular ``__init__`` method.

        See :py:meth:`__init__` for more insight.

        :param data: raw DynamoDB request data.

        :return: fully initialized :py:class:`Table` instance
        """
        hash_key = PrimaryKey.from_dict(data[u'KeySchema'][u'HashKeyElement'])
        range_key = None
        if u'RangeKeyElement' in data[u'KeySchema']:
            range_key = PrimaryKey.from_dict(data[u'KeySchema'][u'RangeKeyElement'])

        return cls( data[u'TableName'],
                    data[u'ProvisionedThroughput'][u'ReadCapacityUnits'],
                    data[u'ProvisionedThroughput'][u'WriteCapacityUnits'],
                    hash_key,
                    range_key,
                  )

    def get_size(self):
        """
        Compute the whole table size using the same rules as the real DynamoDB.
        Actual memory usage in ddbmock will be much higher due to dict and Python
        overheadd.

        .. note:: Real DynamoDB updates this result every 6 hours or so while this is an "on demand" call.

        :return: cumulated size of all items following DynamoDB size computation.
        """
        # TODO: update size only every 6 hours
        size = 0

        for item in self.store:
            size += item.get_size().with_indexing_overhead()

        return size

    def to_dict(self, verbose=True):
        """
        Serialize this table to DynamoDB compatible format. Every fields are
        realistic, including the ``TableSizeBytes`` which relies on
        :py:meth:`get_size.`

        Some DynamoDB requests only send a minimal version of Table metadata. to
        reproduce this behavior, just set ``verbose`` to ``False``.

        :param verbose: Set to ``False`` to skip table size computation.

        :return: Serialized version of table metadata compatible with DynamoDB API syntax.
        """

        ret = {
            "CreationDateTime": self.creation_time,
            "KeySchema": {
                "HashKeyElement": self.hash_key.to_dict(),
            },
            "ProvisionedThroughput": {
                "ReadCapacityUnits": self.rt,
                "WriteCapacityUnits": self.wt,
            },
            "TableName": self.name,
            "TableStatus": self.status
        }

        if verbose:
            ret[u'ItemCount'] = self.count
            ret[u'TableSizeBytes'] = self.get_size()

        if self.last_increase_time:
            ret[u'ProvisionedThroughput'][u'LastIncreaseDateTime'] = self.last_increase_time
        if self.last_decrease_time:
            ret[u'ProvisionedThroughput'][u'LastDecreaseDateTime'] = self.last_decrease_time

        if self.range_key is not None:
            ret[u'KeySchema'][u'RangeKeyElement'] = self.range_key.to_dict()

        return ret

    # these 2 functions helps to pickle table schema (If store supports it)
    # please note that only the schema is persisted, not the state
    def __getstate__(self):
        return (
            self.name,
            self.rt,
            self.wt,
            self.hash_key,
            self.range_key,
            'ACTIVE',
        )

    def __setstate__(self, state):
        self.__init__(*state)  # quick and dirty (tm), but it does the job



# -*- coding: utf-8 -*-

import unittest, mock, json, time

NOW = time.time()
NOW2 = time.time() + 42*1000

TABLE_NAME = 'Table-1'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HEADERS = {
    'x-amz-target': 'dynamodb_20111205.UpdateTable',
    'content-type': 'application/x-amz-json-1.0',
}

# Goal here is not to test the full API, this is done by the Boto tests
class TestUpdateTable(unittest.TestCase):
    @mock.patch("ddbmock.database.table.time")  # Brrr
    def setUp(self, m_time):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        from ddbmock import main
        app = main({})
        from webtest import TestApp
        self.app = TestApp(app)

        m_time.time.return_value = NOW

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)
        t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key, status="ACTIVE")
        dynamodb.data[TABLE_NAME] = t1


    def tearDown(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    @mock.patch("ddbmock.database.table.time")
    def test_update(self, m_time):
        from ddbmock.database.db import dynamodb

        m_time.time.return_value = NOW2

        self.maxDiff = None

        request = {
            "TableName": TABLE_NAME,
            "ProvisionedThroughput": {
                "ReadCapacityUnits": TABLE_RT2,
                "WriteCapacityUnits": TABLE_WT2,
            },
        }

        expected = {
            u'TableDescription': {
                u'CreationDateTime': NOW,
                u'ItemCount': 0,
                u'KeySchema': {
                    u'HashKeyElement': {u'AttributeName': u'hash_key', u'AttributeType': u'N'},
                    u'RangeKeyElement': {u'AttributeName': u'range_key', u'AttributeType': u'S'},
                },
                u'ProvisionedThroughput': {
                    u'LastDecreaseDateTime': NOW2,
                    u'ReadCapacityUnits': TABLE_RT2,
                    u'WriteCapacityUnits': TABLE_WT2,
                },
                u'TableName': TABLE_NAME,
                u'TableSizeBytes': 0,
                u'TableStatus': u'UPDATING'
            }
        }

        # Protocol check
        res = self.app.post_json('/', request, HEADERS, status=200)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])

        # Live data check
        data = dynamodb.data
        assert TABLE_NAME in data
        table = data[TABLE_NAME]

        self.assertEqual(TABLE_RT2, table.rt)
        self.assertEqual(TABLE_WT2, table.wt)



# -*- coding: utf-8 -*-

import unittest, json

TABLE_NAME = 'Table-HR'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE = u'123'
RK_VALUE = u'Decode this data if you are a coder'

HK = {TABLE_HK_TYPE: HK_VALUE}
RK = {TABLE_RK_TYPE: RK_VALUE}

ITEM = {
    TABLE_HK_NAME: HK,
    TABLE_RK_NAME: RK,
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}

HEADERS = {
    'x-amz-target': 'dynamodb_20111205.GetItem',
    'content-type': 'application/x-amz-json-1.0',
}

# Goal here is not to test the full API, this is done by the Boto tests
class TestGetItem(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        from ddbmock.database.db import dynamodb
        from ddbmock import main
        app = main({})
        from webtest import TestApp
        self.app = TestApp(app)
        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        self.t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key)
        dynamodb.data[TABLE_NAME]  = self.t1
        self.t1.put(ITEM, {})

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    def test_get_hr_attr_to_get(self):
        from ddbmock.database.db import dynamodb

        request = {
            "TableName": TABLE_NAME,
            "Key": {
                "HashKeyElement":  HK,
                "RangeKeyElement": RK,
            },
            "AttributesToGet": ["relevant_data"],
        }

        expected = {
            u'ConsumedCapacityUnits': 0.5,
            u'Item': {
                u'relevant_data': ITEM[u'relevant_data'],
            }
        }


        # Protocol check
        res = self.app.post_json('/', request, HEADERS, status=200)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])




# -*- coding: utf-8 -*-

from ddbmock.database import dynamodb

def load_table(func):
    def loader(post, *args):
        name = post[u'TableName']
        table = dynamodb.get_table(name)

        return func(post, table, *args)
    return loader




# -*- coding: utf-8 -*-

class DDBError(Exception):
    def to_dict(self):
        name = type(self).__name__
        msg = self.args[0] if self.args else ""

        return {
            "__type": "com.amazonaws.dynamodb.v20111205#{}".format(name),
            "message": msg,
        }

class BadRequest(DDBError):  status=400; status_str='Bad Request'
class ServerError(DDBError): status=500; status_str='InternalServerError'

class AccessDeniedException(BadRequest): pass
class ConditionalCheckFailedException(BadRequest): pass
class IncompleteSignatureException(BadRequest): pass
class LimitExceededException(BadRequest): pass
class MissingAuthenticationTokenException(BadRequest): pass
class ProvisionedThroughputExceededException(BadRequest): pass
class ResourceInUseException(BadRequest): pass
class ResourceNotFoundException(BadRequest): pass
class ThrottlingException(BadRequest): pass
class ValidationException(BadRequest): pass
class InternalFailure(ServerError): pass
class InternalServerError(ServerError): pass
class ServiceUnavailableException(ServerError): pass



# -*- coding: utf-8 -*-

import sys
from decimal import Decimal

# Types and syntax are already checked
# syntax is expected allright cf input validators

# This module is designed to be used reflexively by ``Scan`` and ``Query``
# implementation through ``getattr``. This helps to keep the code lean in these
# place while allowing to add new operators easily as Amazon implements new ones.
# List of valid Operators are specified in Onctuous validators for each API calls.

# Helpers

def _coerce(typename, value):
    if typename == 'S' or typename == 'B':
        return value
    if typename == 'SS' or typename == 'BS':
        return set(value)
    if typename == 'N':
        return Decimal(value)
    if typename == 'NS':
        return set(map(Decimal, value))

def _parse_elem(elem):
    typename, value = elem.items()[0]
    value = _coerce(typename, value)
    return typename, value


# Common comparison operators

TYPEMSG1 = "Types '{}' is not compatible in test {}"
TYPEMSG2 = "Types '{}' and '{}' are not compatible in test {}"
TYPEMSG3 = "Types '{}', '{} and '{}' are not compatible in test {}"

def eq(target, rule):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype, rvalue = _parse_elem(rule)
    if itype != rtype: raise TypeError(TYPEMSG2.format(itype, rtype, __name__))
    return ivalue == rvalue

def le(target, rule):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype, rvalue = _parse_elem(rule)
    if itype != rtype: raise TypeError(TYPEMSG2.format(itype, rtype, __name__))
    return ivalue <= rvalue

def lt(target, rule):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype, rvalue = _parse_elem(rule)
    if itype != rtype: raise TypeError(TYPEMSG2.format(itype, rtype, __name__))
    return ivalue < rvalue

def ge(target, rule):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype, rvalue = _parse_elem(rule)
    if itype != rtype: raise TypeError(TYPEMSG2.format(itype, rtype, __name__))
    return ivalue >= rvalue

def gt(target, rule):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype, rvalue = _parse_elem(rule)
    if itype != rtype: raise TypeError(TYPEMSG2.format(itype, rtype, __name__))
    return ivalue > rvalue

def begins_with(target, rule):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype, rvalue = _parse_elem(rule)
    if itype != rtype: raise TypeError(TYPEMSG2.format(itype, rtype, __name__))
    return ivalue.startswith(rvalue)

def between(target, rule1, rule2):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype1, rvalue1 = _parse_elem(rule1)
    rtype2, rvalue2 = _parse_elem(rule2)
    if not (itype == rtype1 == rtype2):
        raise TypeError(TYPEMSG3.format(itype, rtype1, rtype2, __name__))
    return ivalue >= rvalue1 and ivalue <= rvalue2

# Comparisons specific to Scan

def null(target):
    return target is None

def not_null(target):
    return not null(target)

def contains(target, rule):
    if target is None: return False
    itype, ivalue = _parse_elem(target)
    rtype, rvalue = _parse_elem(rule)

    if itype[0] != rtype or itype == u"N":
        raise TypeError(TYPEMSG2.format(itype, rtype, __name__))
    return rvalue in ivalue

def not_contains(target, rule):
    return not contains(target, rule)

def in_test(target, *rules):
    if target is None: return False
    itype, ivalue = _parse_elem(target)

    if len(itype) != 1:
        raise TypeError(TYPEMSG1.format(itype, __name__))

    return target in rules

# workaround to use a function with a built-in name
setattr(sys.modules[__name__], 'in', in_test)



# -*- coding: utf-8 -*-
import sys

# Just, skip this unless "--no-skip"
if '--no-skip' in sys.argv:
    import os, imp, boto, unittest
    from ddbmock import connect_boto_patch, clean_boto_patch

    # Load boto dir
    boto_dir = os.path.realpath(boto.__file__+'/../../')

    # Plug patch
    connect_boto_patch()

    # Import boto test modules, pretending they are belonging to us. see hack in tests.unit.__init__
    layer1_file, _, desc1 = imp.find_module('test_layer1', [boto_dir+'/tests/integration/dynamodb'])
    layer2_file, _, desc2 = imp.find_module('test_layer2', [boto_dir+'/tests/integration/dynamodb'])
    layer1_module = imp.load_module('test_layer1', layer1_file, boto_dir, desc1)
    layer2_module = imp.load_module('test_layer2', layer2_file, boto_dir, desc2)

    # Finally, import the tests class, that's enough to let the magic happen !
    DynamoDBLayer1Test = layer1_module.DynamoDBLayer1Test
    DynamoDBLayer2Test = layer2_module.DynamoDBLayer2Test

    # clean up
    layer1_file.close()
    layer2_file.close()
    clean_boto_patch()
else:
    print "Boto Integration tests are slow hence disabled by default, enable them with '--no-skip'"




# -*- coding: utf-8 -*-

from .types import (
    table_name, Required, item_schema, consistent_read, limit, scan_index_forward,
    attributes_to_get_schema, key_field_value, range_key_condition, Boolean,
    get_key_schema)

post = {
    u'TableName': table_name,
    u'HashKeyValue': key_field_value,
    Required(u'RangeKeyCondition', None): range_key_condition,
    Required(u'ScanIndexForward', True): scan_index_forward,
    Required(u'Count', False): Boolean,
    Required(u'Limit', None): limit,
    Required(u'ExclusiveStartKey', None): get_key_schema,
    Required(u'AttributesToGet', []): attributes_to_get_schema,
    Required(u'ConsistentRead', False): consistent_read,  #FIXME: handle default
}



# -*- coding: utf-8 -*-

from . import load_table
from ddbmock.utils import push_read_throughput

@load_table
def get_item(post, table):
    base_capacity = 1 if post[u'ConsistentRead'] else 0.5
    item = table.get(post[u'Key'], post[u'AttributesToGet'])

    if item is not None:
        capacity = base_capacity*item.get_size().as_units()
        push_read_throughput(table.name, capacity)
        return {
            "ConsumedCapacityUnits": capacity,
            "Item": item,
        }
    else:
        push_read_throughput(table.name, base_capacity)
        return {
            "ConsumedCapacityUnits": base_capacity,
        }



# -*- coding: utf-8 -*-

# Very bad import statement
from onctuous import *
from decimal import Decimal
from ddbmock import config
import re

# custom validator for integers
def Precision(min_exp=None, max_exp=None, max_digits=None, msg=None):
    def f(v):
        d = Decimal(v).as_tuple()
        if max_digits is not None and len(d.digits) > max_digits:
            raise Invalid(msg or ('{} has {} digits but maximum is {}'.format(v, len(d.digits), max_digits)))
        if min_exp is not None and min_exp > d.exponent:
            raise Invalid(msg or ('{} exponent is smaller than minimum {}'.format(v, min_exp)))
        if max_exp is not None and max_exp < d.exponent:
            raise Invalid(msg or ('{} exponent is bigger than maximum {}'.format(v, max_exp)))
        return v
    return f

# Elementary types

base_64 = re.compile(r'^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$')

table_name = All(
    unicode,
    Length(min=3, max=255, msg="Table name Length must be between 3 and 255"),
    Match(re.compile(r'^[a-zA-Z0-9\-_\.]*$'), msg="Table name may only contain alphanumeric chars and '-' '_' '.'"),
)

table_page = All(
    int,
    InRange(1, 100, msg="Table pages can only contain from 1 to 100 table"),
)

key_name = All(
    unicode,
    Length(min=1, max=255, msg="Key lenght must be between 1 and 255"),
)

primary_key_type = All(
    unicode,
    Any(u'S', u'N', u'B', msg="Primary key type may only be one of 'S', 'N' or 'B'"),
)

throughtput = All(
    int,
    InRange(config.MIN_TP, config.MAX_TP, msg="throughtput value must be between {} and {}".format(config.MIN_TP, config.MAX_TP)),
)

return_values = All(
    Any(u'NONE', u'ALL_OLD', msg="If specified, return value may only be one of 'NONE' or 'ALL_OLD'"),
)

return_values_all = All(
    Any(u'NONE', u'ALL_OLD', u'UPDATED_OLD', u'ALL_NEW', u'UPDATED_NEW', msg="If specified, return value may only be one of 'ALL_OLD', 'UPDATED_OLD', 'ALL_NEW' or 'UPDATED_NEW'"),
)

consistent_read = All(
    Boolean(msg="Consistent_read parameter must be a Boolean"),
)

limit = All(
    int,
    InRange(min=1, msg="Limit parameter must be a positive integer"),
)

scan_index_forward = All(
    Boolean(msg="ScanIndexForward must be either True either False"),
)

# DynamoDB data types

field_name = unicode

field_number_value = All(
    Precision(min_exp=-128, max_exp=126, max_digits=38),
)

field_string_value = All(unicode, Length(min=1, msg="String fields can not be empty"))

field_binary_value = All(
    unicode,
    Length(min=1, msg="Binary data fields can not be empty"),
    Match(base_64, msg="Binary data must be base64 encoded"),

)

field_number_set_value = All(Length(min=1, msg="A set can not be empty"), [field_number_value])
field_string_set_value = All(Length(min=1, msg="A set can not be empty"), [field_string_value])
field_binary_set_value = All(Length(min=1, msg="A set can not be empty"), [field_binary_value])

# complex types

provisioned_throughtput = {
    u'ReadCapacityUnits':  throughtput,
    u'WriteCapacityUnits': throughtput,
}

primary_key = {
    u'AttributeName': key_name,
    u'AttributeType': primary_key_type,
}

table_key_schema = {
    u'HashKeyElement': primary_key,
    Optional(u'RangeKeyElement'): primary_key,
}

# Fixme: max 1 item
simple_field_value = {
    Optional(u'N'): field_number_value,
    Optional(u'S'): field_string_value,
    Optional(u'B'): field_binary_value,
}

set_field_value = {
    Optional(u'NS'): field_number_set_value,
    Optional(u'SS'): field_string_set_value,
    Optional(u'BS'): field_binary_set_value,
}

key_field_value = simple_field_value
field_value = simple_field_value.copy()
field_value.update(set_field_value)

single_str_num_bin_list = All(Length(min=1, max=1), [simple_field_value])
double_str_num_bin_list = All(Length(min=2, max=2), [simple_field_value])
single_str_bin_list = All(Length(min=1, max=1), [{
    Optional(u'S'): field_string_value,
    Optional(u'B'): field_binary_value,
}])

item_schema = {
    Required(field_name): field_value,
}

get_key_schema = {
    u'HashKeyElement': key_field_value,
    Optional(u'RangeKeyElement'): key_field_value,
}

attributes_to_get_schema = All(
    [field_name],
)

expected_schema = {
    field_name: Any(
        {"Exists": IsFalse()},
        {"Exists": IsTrue(), "Value": field_value}, # pas sur de la vraie syntaxe :/
        {"Value": field_value},
    )
}

# Tss, y'a pas idee de faire des API aussi tordu...
# catches only half of the validity conditions :/
update_action_schema = Any(
    {"Value": field_value, Required("Action", "PUT"): Any("PUT", "ADD")},
    {"Value": set_field_value, "Action": "DELETE"},
    {"Action": "DELETE"},
)

attribute_update_schema = {
    field_name: update_action_schema
}

# Conditions supported by Query
range_key_condition = Any(
    {
        u"ComparisonOperator": Any(u"EQ", u"GT", u"GE", u"LT", u"LE"),
        u"AttributeValueList": single_str_num_bin_list,
    },{
        u"ComparisonOperator": u"BETWEEN",
        u"AttributeValueList": double_str_num_bin_list,
    },{
        u"ComparisonOperator": u"BEGINS_WITH",
        u"AttributeValueList": single_str_bin_list,
    },
)

# Conditions supported by Scan
scan_condition = Any(
    {
        u"ComparisonOperator": Any(u"EQ", u"GT", u"GE", u"LT", u"LE"),
        u"AttributeValueList": single_str_num_bin_list,
    },{
        u"ComparisonOperator": u"BETWEEN",
        u"AttributeValueList": double_str_num_bin_list,
    },{
        u"ComparisonOperator": u"BEGINS_WITH",
        u"AttributeValueList": single_str_bin_list,
    },{
        u"ComparisonOperator": Any(u"NULL", u"NOT_NULL"),
    },{
        u"ComparisonOperator": Any(u"CONTAINS", u"NOT_CONTAINS"),
        u"AttributeValueList": single_str_num_bin_list,
    },{
        u"ComparisonOperator": u"IN",
        u"AttributeValueList": [simple_field_value],
    },
)

# Scan filter
scan_filter = {
    field_name: scan_condition,
}



# -*- coding: utf-8 -*-

from . import load_table
from ddbmock.utils import push_write_throughput

@load_table
def put_item(post, table):
    old, new = table.put(post[u'Item'], post[u'Expected'])
    capacity = max(old.get_size().as_units(), new.get_size().as_units())

    push_write_throughput(table.name, capacity)

    ret = {
        "ConsumedCapacityUnits": capacity,
    }

    if post[u'ReturnValues'] == "ALL_OLD":
        ret["Attributes"] = old

    return ret



# -*- coding: utf-8 -*-

from . import load_table
from ddbmock.utils import push_write_throughput

@load_table
def update_item(post, table):
    old, new = table.update_item(
            post[u'Key'],
            post[u'AttributeUpdates'],
            post[u'Expected'],
    )

    capacity = max(old.get_size().as_units(), new.get_size().as_units())
    push_write_throughput(table.name, capacity)
    ret = {"ConsumedCapacityUnits": capacity}

    if post[u'ReturnValues'] == "ALL_OLD":
        ret["Attributes"] = old
    elif post[u'ReturnValues'] == "ALL_NEW":
        ret["Attributes"] = new
    elif post[u'ReturnValues'] == "UPDATED_OLD":
        ret["Attributes"] = old - new
    elif post[u'ReturnValues'] == "UPDATED_NEW":
        ret["Attributes"] = new - old

    return ret



# -*- coding: utf-8 -*-

import unittest, json

TABLE_NAME1 = 'Table-1'
TABLE_NAME2 = 'Table-2'

HEADERS = {
    'x-amz-target': 'dynamodb_20111205.ListTables',
    'content-type': 'application/x-amz-json-1.0',
}

class TestListTables(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        from ddbmock import main
        app = main({})
        from webtest import TestApp
        self.app = TestApp(app)

        dynamodb.hard_reset()

        hash_key = PrimaryKey('hash_key', 'N')
        range_key = PrimaryKey('range_key', 'S')

        t1 = Table(TABLE_NAME1, 10, 10, hash_key, range_key)
        t2 = Table(TABLE_NAME2, 10, 10, hash_key, range_key)

        dynamodb.data[TABLE_NAME1] = t1
        dynamodb.data[TABLE_NAME2] = t2

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    def test_list_tables(self):
        from ddbmock import connect_boto_patch
        db = connect_boto_patch()

        request = {}

        expected = {
            "TableNames": [TABLE_NAME1, TABLE_NAME2],
        }

        res = self.app.post_json('/', request, HEADERS, status=200)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])




# -*- coding: utf-8 -*-

from ..item import Item
from ddbmock import config
import sqlite3, cPickle as pickle

# I know, using global "variable" for this kind of state *is* bad. But it helps
# keeping execution times to a sane value. In particular, this allows to use
# in-memory version of sqlite
conn = sqlite3.connect(config.STORAGE_SQLITE_FILE)

class Store(object):
    def __init__(self, name):
        """
        Initialize the sqlite store

        By contract, we know the table name will only contain alphanum chars,
        '_', '.' or '-' so that this is ~ safe

        :param name: Table name.
        """
        conn.execute('''CREATE TABLE IF NOT EXISTS `{}` (
          `hash_key` blob NOT NULL,
          `range_key` blob NOT NULL,
          `data` blob NOT NULL,
          PRIMARY KEY (`hash_key`,`range_key`)
        );'''.format(name))
        conn.commit()

        self.name = name

    def truncate(self):
        """
        Perform a full table cleanup. Might be a good idea in tests :)
        """
        conn.execute('DELETE FROM `{}`'.format(self.name))
        conn.commit()

    def _get_by_hash_range(self, hash_key, range_key):
        request = conn.execute('''SELECT `data` FROM `{}`
                                    WHERE `hash_key`=? AND `range_key`=?'''
                                    .format(self.name),
                                    (hash_key, range_key))
        item = request.fetchone()
        if item is None:
            raise KeyError("No item found at ({}, {})".format(hash_key, range_key))

        return pickle.loads(str(item[0]))

    def _get_by_hash(self, hash_key):
        items = conn.execute('''SELECT * FROM `{}`
                                    WHERE `hash_key`=? '''.format(self.name),
                                    (hash_key, ))
        ret = {item[1]:pickle.loads(str(item[2])) for item in items}

        if not ret:
            raise KeyError("No item found at hash_key={}".format(hash_key))
        return ret

    def __getitem__(self, (hash_key, range_key)):
        """
        Get item at (``hash_key``, ``range_key``) or the dict at ``hash_key`` if
        ``range_key``  is None.

        :param key: (``hash_key``, ``range_key``) Tuple. If ``range_key`` is None, all keys under ``hash_key`` are returned
        :return: Item or item dict

        :raise: KeyError
        """

        if range_key is None:
            return self._get_by_hash(hash_key)
        return self._get_by_hash_range(hash_key, range_key)

    def __setitem__(self, (hash_key, range_key), item):
        """
        Set the item at (``hash_key``, ``range_key``). Both keys must be
        defined and valid. By convention, ``range_key`` may be ``False`` to
        indicate a ``hash_key`` only key.

        :param key: (``hash_key``, ``range_key``) Tuple.
        :param item: the actual ``Item`` data structure to store
        """
        db_item = buffer(pickle.dumps(item, 2))
        conn.execute('''INSERT OR REPLACE INTO `{}` (`hash_key`,`range_key`, `data`)
                     VALUES (?, ?, ?)'''.format(self.name),
                     (hash_key, range_key, db_item))
        conn.commit()

    def __delitem__(self, (hash_key, range_key)):
        """
        Delete item at key (``hash_key``, ``range_key``)

        :raises: KeyError if not found
        """
        conn.execute('DELETE FROM `{}` WHERE `hash_key`=? AND `range_key`=?'
                          .format(self.name), (hash_key, range_key))

    def __iter__(self):
        """
        Iterate all over the table, abstracting the ``hash_key`` and
        ``range_key`` complexity. Mostly used for ``Scan`` implementation.
        """
        items = conn.execute('SELECT `data` FROM `{}`'.format(self.name))
        for item in items:
            yield pickle.loads(str(item[0]))



# -*- coding: utf-8 -*-

import unittest, mock

# This module both validates schema persitence/reloading and that sqlite backend is
# actually working

TABLE_NAME = "tabloid"

TABLE_RT = 45
TABLE_WT = 123

HASH_KEY = {"AttributeName":"hash_key","AttributeType":"N"}
RANGE_KEY = {"AttributeName":"range_key","AttributeType":"S"}

class TestDBSchemaPersist(unittest.TestCase):
    @mock.patch('ddbmock.database.db.Store')
    def test_db_schema_persistence(self, m_store):
        from ddbmock.database.db import DynamoDB
        from ddbmock.database import storage
        from ddbmock.database.storage.sqlite import Store

        data = {
            "TableName": TABLE_NAME,
            "KeySchema": {
                "HashKeyElement": HASH_KEY,
                "RangeKeyElement": RANGE_KEY,
            },
            "ProvisionedThroughput": {
                "ReadCapacityUnits": TABLE_RT,
                "WriteCapacityUnits": TABLE_WT,
            }
        }

        m_store.return_value = Store("whatever")

        # reset internal state + force SQLite
        old_internal_state = DynamoDB._shared_data
        DynamoDB._shared_data = {'data': {}, 'store': None}

        # data creation round
        dynamodb = DynamoDB()
        dynamodb.create_table(TABLE_NAME, data)

        # reset internal state (again) to simulate reload
        DynamoDB._shared_data = {'data': {}, 'store': None}
        dynamodb = DynamoDB()

        table = dynamodb.data[TABLE_NAME]
        self.assertEqual(TABLE_NAME, table.name)
        self.assertEqual(TABLE_RT, table.rt)
        self.assertEqual(TABLE_WT, table.wt)

        DynamoDB._shared_data = old_internal_state



# -*- coding: utf-8 -*-

from time import time, ctime
from collections import namedtuple
import atexit, logging

null_logger = logging.getLogger(__name__)
null_logger.addHandler(logging.NullHandler())


def average(data):
    return sum(data)/len(data)

class Stat(object):
    def __init__(self, name, resolution_interval=1, aggregation_interval=5*60, logger=null_logger):
        """
        :param name: name of the measured parameter
        :param resolution_interval: all data accumulated in this slot are part of the same point
        :param aggregation_interval: aggregates data on this period. Must be bigger than ``resolution_interval``
        """

        # Load params
        self.name=name
        self.resolution_interval = resolution_interval
        self.aggregation_interval = aggregation_interval
        self.log = logger

        # Set internal state
        self.current_point_time = int(time())
        self.current_point_list = []
        self.current_point_value = 0
        self.last_aggregation_time = self.current_point_time

        # goodbye
        atexit.register(self.flush)

    def _macro_aggregate(self):
        """Perform aggregation every aggregation_interval"""
        # aggregate
        points = self.current_point_list

        interval = (self.current_point_time - self.last_aggregation_time) / 60.0
        self.log.info("%s: interval=%s min=%s max=%s average=%s",
                       self.name,
                       round(interval),
                       min(points),
                       max(points),
                       average(points))

        #reset
        self.current_point_list = []
        self.last_aggregation_time = int(time())

    def _aggregate(self):
        """Trigger aggregation and reset current data"""
        # aggregate
        self.current_point_list.append(self.current_point_value)

        # reset
        self.current_point_value = 0
        self.current_point_time = int(time())


    def push(self, value):
        """Push a data point

        :param value: value to push
        """
        current_time = int(time())

        # aggregate ?
        if self.current_point_time + self.current_point_time <= current_time:
            self._aggregate()
        if self.last_aggregation_time + self.aggregation_interval <= current_time:
            self._macro_aggregate()

        # update value
        self.current_point_value += value

    def flush(self):
        """Force all aggregations, reset internal state and print the results
        """

        self._aggregate()
        self._macro_aggregate()



# -*- coding: utf-8 -*-

from importlib import import_module
from ddbmock.utils import req_logger
from ddbmock.errors import InternalFailure
from ddbmock.validators import dynamodb_api_validate
import re, itertools

request_counter = itertools.count()  # atomic counter

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def action_to_route(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()

def router(action, post):
    request_id = str(request_counter.next())  # register the id in the post for it to be accessible in case of exception
    post['request_id'] = request_id
    req_logger.debug("request_id=%s action=%s body=%s", request_id, action, post)

    # Find route
    try:
        target = action_to_route(action)
        mod = import_module('ddbmock.operations.{}'.format(target))
        func = getattr(mod, target)
    except ImportError:
        req_logger.error('request_id=%s action=%s No such action', request_id, action)
        raise InternalFailure("Method: {} does not exist".format(action))

    # Validate the input
    try:
        post = dynamodb_api_validate(target, post)
    except Exception as e:
        req_logger.error('request_id=%s action=%s exception=%s body=%s', request_id, action, type(e).__name__, str(e.args))
        raise

    # Run request and translate engine errors to DynamoDB errors
    try:
        answer = func(post)
        #req_logger.debug("request_id=%s action=%s answer=%s", post['request_id'], action, answer)
        return answer
    except (TypeError, ValueError, KeyError) as e:
        req_logger.error('request_id=%s action=%s exception=%s body=%s', request_id, action, type(e).__name__, str(e.args))
        raise InternalFailure("{}: {}".format(type(e).__name__, str(e.args)))



# -*- coding: utf-8 -*-

from .types import table_name, Required, item_schema, get_key_schema, consistent_read, attributes_to_get_schema

post = {
    u'TableName': table_name,
    u'Key': get_key_schema,
    Required(u'AttributesToGet', []): attributes_to_get_schema,
    Required(u'ConsistentRead', False): consistent_read,
}



# -*- coding: utf-8 -*-

import unittest
import boto

TABLE_NAME = 'Table-1'
TABLE_NAME_404= 'Waldo'
TABLE_RT = 45
TABLE_WT = 123
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE = u'123'
RK_VALUE = u'Decode this data if you are a coder'
RK_VALUE2 = u'Decode this data if you are a coder (or not)'


ITEM1 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}
ITEM2 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE2},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}

class TestDescribeTables(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key, status='ACTIVE')

        t1.put(ITEM1, {})
        t1.put(ITEM2, {})

        dynamodb.data[TABLE_NAME] = t1

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_describe_table(self):
        from ddbmock import connect_boto_patch
        db = connect_boto_patch()

        table = db.get_table(TABLE_NAME)

        self.assertEqual(TABLE_NAME, table.name)
        self.assertEqual(421, table.size_bytes)
        self.assertEqual(TABLE_RT, table.read_units)
        self.assertEqual(TABLE_WT, table.write_units)
        self.assertEqual(u'ACTIVE', table.status)
        self.assertEqual(TABLE_HK_NAME, table.schema.hash_key_name)
        self.assertEqual(TABLE_HK_TYPE, table.schema.hash_key_type)
        self.assertEqual(TABLE_RK_NAME, table.schema.range_key_name)
        self.assertEqual(TABLE_RK_TYPE, table.schema.range_key_type)

    def describe_table_404(self):
        from ddbmock import connect_boto_patch
        db = connect_boto_patch()

        self.assertRaises(DDBValidationErr, db.get_table,
            TABLE_NAME_404,
        )




# -*- coding: utf-8 -*-

from .types import (
    table_name, Required, item_schema, consistent_read, limit, scan_filter,
    attributes_to_get_schema, key_field_value, Boolean, get_key_schema)

post = {
    u'TableName': table_name,
    Required(u'ScanFilter', {}): scan_filter,
    Required(u'Count', False): Boolean,
    Required(u'Limit', None): limit,
    Required(u'ExclusiveStartKey', None): get_key_schema,
    Required(u'AttributesToGet', []): attributes_to_get_schema,
}



# -*- coding: utf-8 -*-

from .types import (table_name, get_key_schema, return_values_all,
                    expected_schema, attribute_update_schema, Required)

post = {
    u'TableName': table_name,
    u'Key': get_key_schema,
    u'AttributeUpdates': attribute_update_schema,
    Required(u'Expected', {}): expected_schema,
    Required(u'ReturnValues', u'NONE'): return_values_all,
}



# -*- coding: utf-8 -*-

from collections import defaultdict

class Store(object):
    def __init__(self, name):
        """
        Initialize the in-memory store

        :param name: Table name.
        """
        self.name = name
        self.data = defaultdict(dict)

    def truncate(self):
        """Perform a full table cleanup. Might be a good idea in tests :)"""
        self.data = defaultdict(dict)

    def __getitem__(self, (hash_key, range_key)):
        """
        Get item at (``hash_key``, ``range_key``) or the dict at ``hash_key`` if
        ``range_key``  is None.

        :param key: (``hash_key``, ``range_key``) Tuple. If ``range_key`` is None, all keys under ``hash_key`` are returned
        :return: Item or item dict

        :raise: KeyError
        """

        if hash_key not in self.data:
            raise KeyError('hash_key={} not found'.format(hash_key))
        if range_key is None:
            return self.data[hash_key]
        if range_key not in self.data[hash_key]:
            raise KeyError('range_key={} not found'.format(range_key))
        return self.data[hash_key][range_key]

    def __setitem__(self, (hash_key, range_key), item):
        """
        Set the item at (``hash_key``, ``range_key``). Both keys must be
        defined and valid. By convention, ``range_key`` may be ``False`` to
        indicate a ``hash_key`` only key.

        :param key: (``hash_key``, ``range_key``) Tuple.
        :param item: the actual ``Item`` data structure to store
        """

        self.data[hash_key][range_key] = item

    def __delitem__(self, (hash_key, range_key)):
        """
        Delete item at key (``hash_key``, ``range_key``)

        :raises: KeyError if not found
        """

        # Let this line throw KeyError if needed. Checks needs to be performed by the caller
        del self.data[hash_key][range_key]

    def __iter__(self):
        """
        Iterate all over the table, abstracting the ``hash_key`` and
        ``range_key`` complexity. Mostly used for ``Scan`` implementation.
        """

        for outer in self.data.values():
            for item in outer.values():
                yield item



# -*- coding: utf-8 -*-

from . import load_table
from ddbmock.utils import push_write_throughput
from ddbmock.errors import ValidationException

@load_table
def query(post, table):
    if post[u'AttributesToGet'] and post[u'Count']:
        raise ValidationException("Can filter fields when only count is requested")

    base_capacity = 1 if post[u'ConsistentRead'] else 0.5

    results = table.query(
        post[u'HashKeyValue'],
        post[u'RangeKeyCondition'],
        post[u'AttributesToGet'],
        post[u'ExclusiveStartKey'],
        not post[u'ScanIndexForward'],
        post[u'Limit'],
    )

    capacity = base_capacity*results.size.as_units()
    push_write_throughput(table.name, capacity)

    ret = {
        "Count": len(results.items),
        "ConsumedCapacityUnits": capacity,
    }

    if results.last_key is not None:
        ret['LastEvaluatedKey'] = results.last_key

    if not post[u'Count']:
        ret[u'Items'] = results.items

    return ret



# -*- coding: utf-8 -*-

from onctuous import Schema, Invalid
from importlib import import_module
from ddbmock.errors import ValidationException

def dynamodb_api_validate(action, post):
    """ Find validator for ``action`` and run it on ``post``. If no validator
        are found, return False.

        :action: name of the route after translation to underscores
        :post: data to validate
        :return: False when no validator found, True on success
        :raises: any onctuous exception
    """
    try:
        mod = import_module('.{}'.format(action), __name__)
        schema = getattr(mod, 'post')
    except (ImportError, AttributeError):
        return post  # Fixme: should log

    # ignore the 'request_id' key but propagate it
    schema['request_id'] = str

    try:
        validate = Schema(schema, required=True)
        return validate(post)
    except Invalid as e:
        raise ValidationException(str(e.errors))



# -*- coding: utf-8 -*-

import unittest
import boto
from decimal import Decimal
from copy import deepcopy as cp

TABLE_NAME = 'Table-HR'
TABLE_NAME2 = 'Table-H'
TABLE_NAME_404 = 'Waldo'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE = u'123'
HK_VALUE2 = u'456'
RK_VALUE = u'Decode this data if you are a coder'

RELEVANT_FIELD = {'S': 'Illyse'}
IRELEVANT_FIELD = {'B': 'WW91IHdpc2ggeW91IGNvdWxkIGNoYW5nZSB5b3VyIGpvYi4uLg=='}

FIELD_NAME = u"relevant_data"
FIELD_NAME_404 = u"ze dummy field name"
FIELD_SET_NAME = u"data list"
FIELD_NUM_NAME = u"counter"

RELEVANT_HUGE_FIELD = {'S': 'a'*64*1024}

ITEM = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE},
    FIELD_NAME: {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
    FIELD_SET_NAME: {u'SS': [u'item1', u'item2', u'item3', u'item4']}
}

ITEM2 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
    FIELD_NUM_NAME: {u'N': u'42'}

}

FIELD_SMALL = {u'S': u'a'}
FIELD_BIG   = {u'S': u'a'*1024}



class TestUpdateItem(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        self.t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key)
        self.t2 = Table(TABLE_NAME2, TABLE_RT, TABLE_WT, hash_key, None)

        dynamodb.data[TABLE_NAME]  = self.t1
        dynamodb.data[TABLE_NAME2] = self.t2

        self.t1.put(cp(ITEM), {})
        self.t2.put(cp(ITEM2), {})

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_update_item_put_hr(self):
        from ddbmock import connect_boto_patch
        db = connect_boto_patch()

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        key2 = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE2},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        # use PUT as default action, champs existant
        db.layer1.update_item(TABLE_NAME, key, {
            'relevant_data': {'Value': RELEVANT_FIELD}  # Move type from 'B' to 'S'
        })
        self.assertEqual(RELEVANT_FIELD, self.t1.store[HK_VALUE, RK_VALUE]['relevant_data'])

        # PUT explicite, champ non existant
        db.layer1.update_item(TABLE_NAME, key, {
            'irelevant_data': {'Action': 'PUT', 'Value': IRELEVANT_FIELD}
        })
        self.assertEqual(RELEVANT_FIELD, self.t1.store[HK_VALUE, RK_VALUE]['relevant_data'])
        self.assertEqual(IRELEVANT_FIELD, self.t1.store[HK_VALUE, RK_VALUE]['irelevant_data'])

        # PUT explicite, item non existant(full item creation)
        db.layer1.update_item(TABLE_NAME, key2, {
            'relevant_data': {'Action': 'PUT', 'Value': RELEVANT_FIELD}
        })
        self.assertEqual({TABLE_HK_TYPE: HK_VALUE2}, self.t1.store[HK_VALUE2, RK_VALUE][TABLE_HK_NAME])
        self.assertEqual({TABLE_RK_TYPE: RK_VALUE}, self.t1.store[HK_VALUE2, RK_VALUE][TABLE_RK_NAME])
        self.assertEqual(RELEVANT_FIELD, self.t1.store[HK_VALUE2, RK_VALUE]['relevant_data'])

    def test_update_item_put_h(self):
        from ddbmock import connect_boto_patch
        db = connect_boto_patch()

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
        }

        key2 = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE2},
        }

        # use PUT as default action, champs existant
        db.layer1.update_item(TABLE_NAME2, key, {
            'relevant_data': {'Value': RELEVANT_FIELD}  # Move type from 'B' to 'S'
        })
        self.assertEqual(RELEVANT_FIELD, self.t2.store[HK_VALUE, False]['relevant_data'])

        # PUT explicite, champ non existant
        db.layer1.update_item(TABLE_NAME2, key, {
            'irelevant_data': {'Action': 'PUT', 'Value': IRELEVANT_FIELD}
        })
        self.assertEqual(RELEVANT_FIELD, self.t2.store[HK_VALUE, False]['relevant_data'])
        self.assertEqual(IRELEVANT_FIELD, self.t2.store[HK_VALUE, False]['irelevant_data'])

        # PUT explicite, item non existant(full item creation)
        db.layer1.update_item(TABLE_NAME2, key2, {
            'relevant_data': {'Action': 'PUT', 'Value': RELEVANT_FIELD}
        })
        self.assertEqual({TABLE_HK_TYPE: HK_VALUE2}, self.t2.store[HK_VALUE2, False][TABLE_HK_NAME])
        self.assertEqual(RELEVANT_FIELD, self.t2.store[HK_VALUE2, False]['relevant_data'])

    def test_put_check_throughput_max_old_new(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        key = {u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE}}

        self.assertEqual(
            {u'ConsumedCapacityUnits': 1},
            db.layer1.update_item(TABLE_NAME2, key,
                                  {FIELD_NAME: {'Action': 'PUT', 'Value': FIELD_SMALL}}),
        )
        self.assertEqual(
            {u'ConsumedCapacityUnits': 2},
            db.layer1.update_item(TABLE_NAME2, key,
                                  {FIELD_NAME: {'Action': 'PUT', 'Value': FIELD_BIG}}),
        )
        self.assertEqual(
            {u'ConsumedCapacityUnits': 2},
            db.layer1.update_item(TABLE_NAME2, key,
                                  {FIELD_NAME: {'Action': 'PUT', 'Value': FIELD_SMALL}}),
        )

    def test_update_item_delete_primary_key_fails(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        self.assertRaises(DynamoDBValidationError,
            db.layer1.update_item,
            TABLE_NAME, key, {TABLE_RK_NAME: {'Action': 'DELETE'}}
        )
        self.assertEqual({TABLE_RK_TYPE: RK_VALUE}, self.t1.store[HK_VALUE, RK_VALUE][TABLE_RK_NAME])

        self.assertRaises(DynamoDBValidationError,
            db.layer1.update_item,
            TABLE_NAME, key, {TABLE_HK_NAME: {'Action': 'DELETE'}}
        )
        self.assertEqual({TABLE_HK_TYPE: HK_VALUE}, self.t1.store[HK_VALUE, RK_VALUE][TABLE_HK_NAME])

    def test_update_item_delete_field_ok(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        db.layer1.update_item(TABLE_NAME, key, {
            FIELD_NAME: {'Action': 'DELETE'},
        })
        self.assertNotIn(FIELD_NAME, self.t1.store[HK_VALUE, RK_VALUE])

        # Attempt to delete non-existing field, do nothing
        db.layer1.update_item(TABLE_NAME, key, {
            FIELD_NAME_404: {'Action': 'DELETE'},
        })

    def test_update_item_delete_field_set_ok(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        expected1 = {u'SS': [u'item1', u'item2', u'item3', u'item4']}
        expected2 = {u'SS': [u'item3', u'item1']}

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        # Can not remove a scalar value, even from a set
        self.assertRaises(DynamoDBValidationError,
            db.layer1.update_item,
            TABLE_NAME, key, {
                FIELD_SET_NAME: {'Action': 'DELETE', u'Value': {u'S': u'item1'}},
            }
        )
        self.assertEqual(expected1, self.t1.store[HK_VALUE, RK_VALUE][FIELD_SET_NAME])

        # remove a couple of existing or not item from the field
        db.layer1.update_item(TABLE_NAME, key, {
            FIELD_SET_NAME: {'Action': 'DELETE', u'Value': {u'SS': [u'item2', u'item4', u'item6']}},
        })
        self.assertEqual(expected2, self.t1.store[HK_VALUE, RK_VALUE][FIELD_SET_NAME])

        # Field shoud disapear (Empty)
        db.layer1.update_item(TABLE_NAME, key, {
            FIELD_SET_NAME: {'Action': 'DELETE', u'Value': {u'SS': [u'item1', u'item3', u'item6']}},
        })
        self.assertNotIn(FIELD_SET_NAME, self.t1.store[HK_VALUE, RK_VALUE])

    def test_update_item_delete_field_set_bad_type(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        expected = {u'SS': [u'item1', u'item2', u'item3', u'item4']}

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        self.assertRaises(DynamoDBValidationError,
            db.layer1.update_item,
            TABLE_NAME, key, {
                FIELD_SET_NAME: {'Action': 'DELETE', u'Value': {u'B': u'item1'}},
            }
        )
        self.assertEqual(expected, self.t1.store[HK_VALUE, RK_VALUE][FIELD_SET_NAME])

        self.assertRaises(DynamoDBValidationError,
            db.layer1.update_item,
            TABLE_NAME, key, {
                FIELD_SET_NAME: {'Action': 'DELETE', u'Value': {u'BS': [u'item2', u'item4', u'item6']}},
            }
        )
        self.assertEqual(expected, self.t1.store[HK_VALUE, RK_VALUE][FIELD_SET_NAME])

    def test_update_item_increment(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
        }

        ADD_VALUE = 2

        expected = {u'N': unicode(42 + ADD_VALUE)}

        # regular increment
        db.layer1.update_item(TABLE_NAME2, key, {
            FIELD_NUM_NAME: {'Action': 'ADD', u'Value': {u'N': unicode(ADD_VALUE)}},
        })
        self.assertEqual(expected, self.t2.store[HK_VALUE, False][FIELD_NUM_NAME])

    def test_update_item_push_to_set_ok(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        expected1 = {u'SS': [u'item1', u'item2', u'item3', u'item4']}
        expected2 = {u'SS': [u'item2', u'item3', u'item1', u'item4', u'item5']}

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        self.assertRaises(DynamoDBValidationError,
            db.layer1.update_item,
            TABLE_NAME, key, {
                FIELD_SET_NAME: {'Action': 'ADD', u'Value': {u'S': u'item5'}},
            }
        )
        self.assertEqual(expected1, self.t1.store[HK_VALUE, RK_VALUE][FIELD_SET_NAME])

        db.layer1.update_item(TABLE_NAME, key, {
            FIELD_SET_NAME: {'Action': 'ADD', u'Value': {u'SS': [u'item5']}},
        })
        self.assertEqual(expected2, self.t1.store[HK_VALUE, RK_VALUE][FIELD_SET_NAME])

    def test_update_item_push_to_non_set_fail(self):
        # sometimes weird black magic types occures in test. these are related
        # to internal "DB" logic. it does not affect real API output at all
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        expected1 = {u'SS': [u'item1', u'item2', u'item3', u'item4']}

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        self.assertRaises(DynamoDBValidationError,
            db.layer1.update_item,
            TABLE_NAME, key, {
                FIELD_NAME: {'Action': 'ADD', u'Value': {u'B': u'item5'}},
            }
        )
        self.assertEqual(expected1, self.t1.store[HK_VALUE, RK_VALUE][FIELD_SET_NAME])

    def test_update_return_all_old(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        key = {u"HashKeyElement": {TABLE_HK_TYPE: HK_VALUE}}
        ADD_VALUE = 1

        db = connect_boto_patch()
        expected = cp(ITEM2)

        # regular increment
        ret = db.layer1.update_item(TABLE_NAME2, key, {
                FIELD_NUM_NAME: {'Action': 'ADD', u'Value': {u'N': unicode(ADD_VALUE)}},
            },
            return_values=u'ALL_OLD',
         )
        self.assertEqual(expected, ret[u'Attributes'])

    def test_update_return_all_new(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        key = {u"HashKeyElement": {TABLE_HK_TYPE: HK_VALUE}}
        ADD_VALUE = 1

        db = connect_boto_patch()
        expected = cp(ITEM2)
        expected[FIELD_NUM_NAME][u'N'] = unicode(int(expected[FIELD_NUM_NAME][u'N']) + ADD_VALUE)

        # regular increment
        ret = db.layer1.update_item(TABLE_NAME2, key, {
                FIELD_NUM_NAME: {'Action': 'ADD', u'Value': {u'N': unicode(ADD_VALUE)}},
            },
            return_values=u'ALL_NEW',
        )
        self.assertEqual(expected, ret[u'Attributes'])

    def test_update_return_updated_old(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        key = {u"HashKeyElement": {TABLE_HK_TYPE: HK_VALUE}}
        ADD_VALUE = 1

        db = connect_boto_patch()
        expected = {FIELD_NUM_NAME: cp(ITEM2[FIELD_NUM_NAME])}

        # regular increment
        ret = db.layer1.update_item(TABLE_NAME2, key, {
                FIELD_NUM_NAME: {'Action': 'ADD', u'Value': {u'N': unicode(ADD_VALUE)}},
            },
            return_values=u'UPDATED_OLD',
         )
        self.assertEqual(expected, ret[u'Attributes'])

    def test_update_return_updated_new(self):
        from ddbmock import connect_boto_patch
        from boto.dynamodb.exceptions import DynamoDBValidationError

        key = {u"HashKeyElement": {TABLE_HK_TYPE: HK_VALUE}}
        ADD_VALUE = 1

        db = connect_boto_patch()
        expected = {FIELD_NUM_NAME: cp(ITEM2[FIELD_NUM_NAME])}
        expected[FIELD_NUM_NAME][u'N'] = unicode(int(expected[FIELD_NUM_NAME][u'N']) + ADD_VALUE)

        # regular increment
        ret = db.layer1.update_item(TABLE_NAME2, key, {
                FIELD_NUM_NAME: {'Action': 'ADD', u'Value': {u'N': unicode(ADD_VALUE)}},
            },
            return_values=u'UPDATED_NEW',
         )
        self.assertEqual(expected, ret[u'Attributes'])

    def test_update_item_put_h_oversized(self):
        from boto.dynamodb.exceptions import DynamoDBValidationError
        from ddbmock import connect_boto_patch
        db = connect_boto_patch()

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
        }

        # PUT explicite, existing field
        self.assertRaisesRegexp(DynamoDBValidationError, 'Items.*smaller.*update',
        db.layer1.update_item, TABLE_NAME2, key, {
            'relevant_data': {'Value': RELEVANT_HUGE_FIELD},
        })
        self.assertEqual(ITEM[FIELD_NAME], self.t2.store[HK_VALUE, False]['relevant_data'])



# -*- coding: utf-8 -*-

import unittest, mock

LOGGER = 'unit-stat-test'
DELAY = 12

class TestUtils(unittest.TestCase):
    @mock.patch('ddbmock.utils.Timer')
    def test_schedule_action_enable(self, m_timer):
        from ddbmock.utils import schedule_action
        from ddbmock import config

        old_delay_param = config.ENABLE_DELAYS
        config.ENABLE_DELAYS = True

        m_cb = mock.Mock()

        schedule_action(DELAY, m_cb, ["args"], {'kwargs':'kwargs'})
        m_timer.assert_called_with(DELAY, m_cb, ["args"], {'kwargs':'kwargs'})
        m_timer.return_value.start.assert_called_with()

        config.ENABLE_DELAYS = old_delay_param

    @mock.patch('ddbmock.utils.Timer')
    def test_schedule_action_disable(self, m_timer):
        from ddbmock.utils import schedule_action
        from ddbmock import config

        old_delay_param = config.ENABLE_DELAYS
        config.ENABLE_DELAYS = False

        m_cb = mock.Mock()

        schedule_action(DELAY, m_cb, ["args"], {'kwargs':'kwargs'})
        m_cb.assert_called_with("args", kwargs='kwargs')

        config.ENABLE_DELAYS = old_delay_param

    def test_average(self):
        from ddbmock.utils.stat import average

        self.assertEqual(2, average([1,2,3]))
        self.assertEqual(2, average([1,2,3,4]))

    @mock.patch('ddbmock.utils.stat.time')
    @mock.patch('ddbmock.utils.stat.atexit')
    def test_aggregate(self, _, m_time):
        from ddbmock.utils.stat import Stat

        s = Stat(LOGGER)
        s.current_point_value = 42

        m_time.return_value = 123.5

        s._aggregate()

        self.assertEqual([42], s.current_point_list)
        self.assertEqual(0, s.current_point_value)
        self.assertEqual(123, s.current_point_time)

    @mock.patch('ddbmock.utils.stat.atexit')
    def test_macro_aggregate(self, _):
        from ddbmock.utils.stat import Stat

        s = Stat(LOGGER)

        s.current_point_list = [-4,12,5000]
        s.last_aggregation_time = 32

        s._macro_aggregate()

        self.assertEqual([], s.current_point_list)

    @mock.patch('ddbmock.utils.stat.time')
    @mock.patch('ddbmock.utils.stat.Stat._aggregate')
    @mock.patch('ddbmock.utils.stat.Stat._macro_aggregate')
    @mock.patch('ddbmock.utils.stat.atexit')
    def test_push(self, _, _m_m_agg, _m_agg, m_time):
        from ddbmock.utils.stat import Stat

        m_time.return_value = 1
        s = Stat(LOGGER)

        # nominal
        s.push(1)
        s.push(2)
        s.push(3)
        self.assertEqual(6, s.current_point_value)
        self.assertFalse(_m_agg.called)
        self.assertFalse(_m_m_agg.called)

        # manual clean (mocks...)
        s.current_point_value = 0

        # change data point
        m_time.return_value = 2.1
        s.push(4)
        self.assertEqual(4, s.current_point_value)
        self.assertTrue(_m_agg.called)
        self.assertFalse(_m_m_agg.called)

        # manual clean (mocks...)
        s.current_point_value = 0

        # change data point
        m_time.return_value = 5*60+1
        s.push(6)
        self.assertEqual(6, s.current_point_value)
        self.assertTrue(_m_agg.called)
        self.assertTrue(_m_m_agg.called)

    @mock.patch('ddbmock.utils.stat.Stat._aggregate')
    @mock.patch('ddbmock.utils.stat.Stat._macro_aggregate')
    @mock.patch('ddbmock.utils.stat.atexit')
    def test_flush(self, _, _m_m_agg, _m_agg):
        from ddbmock.utils.stat import Stat

        s = Stat(LOGGER)
        m_list = mock.MagicMock()
        s.flush()

        self.assertTrue(_m_agg.called)
        self.assertTrue(_m_m_agg.called)



# -*- coding: utf-8 -*-

from .table import Table
from .item import ItemSize
from .storage import Store
from collections import defaultdict
from ddbmock import config
from ddbmock.utils import push_write_throughput, push_read_throughput, schedule_action
from ddbmock.errors import (ResourceNotFoundException,
                            ResourceInUseException,
                            LimitExceededException,
                           )

# All validations are performed on *incomming* data => already done :)

class DynamoDB(object):
    """
    Main database, behaves as a singleton in the sense that all instances share
    the same data.

    If underlying store supports it, all tables schema are persisted at creation
    time to a special table ``~*schema*~`` which is an invalid DynamoDB table name
    so no collisions are to be expected.
    """
    _shared_data = {
        'data': {},
        'store': None,
    }

    def __init__(self):
        """
        When first instanciated, ``__init__`` checks the underlying store for
        potentially persisted tables. If any, it reloads there schema to make
        them available to the application.

        In all other cases, ``__init__`` simply loads the shared state.
        """
        cls = type(self)
        self.__dict__ = cls._shared_data

        # At first instanciation, attempt to reload the database schema
        if self.store is None:
            self.store = Store('~*schema*~')
            for table in self.store:
                self.data[table.name] = table

    def hard_reset(self):
        """
        Reset and drop all tables. If any data was persisted, it will be completely
        lost after a call to this method. I do use in ``tearDown`` of all ddbmock
        tests to avaid any side effect.
        """
        for table in self.data.values():
            table.truncate()

        self.data.clear()
        self.store.truncate()

    def list_tables(self):
        """
        Get a list of all table names.
        """
        return self.data.keys()

    def get_table(self, name):
        """
        Get a handle to :py:class:`ddbmock.database.table.Table` '``name``' is it exists.

        :param name: Name of the table to load.

        :return: :py:class:`ddbmock.database.table.Table` with name '``name``'

        :raises: :py:exc:`ddbmock.errors.ResourceNotFoundException` if the table does not exist.
        """
        if name in self.data:
            return self.data[name]
        raise ResourceNotFoundException("Table {} does not exist".format(name))

    def create_table(self, name, data):
        """
        Create a :py:class:`ddbmock.database.table.Table` named '``name``'' using
        parameters provided in ``data`` if it does not yet exist.

        :param name: Valid table name. No further checks are performed.
        :param data: raw DynamoDB request data.

        :return: A reference to the newly created :py:class:`ddbmock.database.table.Table`

        :raises: :py:exc:`ddbmock.errors.ResourceInUseException` if the table already exists.
        :raises: :py:exc:`ddbmock.errors.LimitExceededException` if more than :py:const:`ddbmock.config.MAX_TABLES` already exist.
        """
        if name in self.data:
            raise ResourceInUseException("Table {} already exists".format(name))
        if len(self.data) >= config.MAX_TABLES:
            raise LimitExceededException("Table limit reached. You can not have more than {} tables simultaneously".format(config.MAX_TABLES))

        self.data[name] = Table.from_dict(data)
        self.store[name, False] = self.data[name]
        return self.data[name]

    def _internal_delete_table(self, table):
        """This is ran only after the timer is exhausted.
        Important note: this function is idempotent. If another table with the
        same name has been created in the mean time, it won't be overwrittem.
        This is the kind of special cases you might encounter in testing
        environenment"""
        name = table.name
        if name in self.data and self.data[name] is table:
            self.data[name].store.truncate()  # FIXME: should be moved in table
            del self.data[name]
            del self.store[name, False]

    def delete_table(self, name):
        """
        Triggers internal "realistic" table deletion. This implies changing
        the status to ``DELETING``. Once :py:const:ddbmock.config.DELAY_DELETING
        has expired :py:meth:_internal_delete_table is called and the table
        is de-referenced from :py:attr:data.

        Since :py:attr:data only holds a reference, the table object might still
        exist at that moment and possibly still handle pending requests. This also
        allows to safely return a handle to the table object.

        :param name: Valid table name.

        :return: A reference to :py:class:`ddbmock.database.table.Table` named ``name``
        """
        if name not in self.data:
            raise ResourceNotFoundException("Table {} does not exist".format(name))

        table = self.data[name]
        table.delete()
        schedule_action(config.DELAY_DELETING, self._internal_delete_table, [table])

        return table

    def get_batch(self, batch):
        """
        Batch processor. Dispatches call to appropriate :py:class:`ddbmock.database.table.Table`
        methods. This is the only low_level API that directly pushes throughput usage.

        :param batch: raw DynamoDB request batch.

        :returns: dict compatible with DynamoDB API

        :raises: :py:exc:`ddbmock.errors.ValidationException` if a ``range_key`` was provided while table has none.
        :raises: :py:exc:`ddbmock.errors.ResourceNotFoundException` if a table does not exist.
        """
        ret = defaultdict(dict)

        for tablename, batch in batch.iteritems():
            base_capacity = 1 if batch[u'ConsistentRead'] else 0.5
            fields = batch[u'AttributesToGet']
            table = self.get_table(tablename)
            units = ItemSize(0)
            items = []
            for key in batch[u'Keys']:
                item = table.get(key, fields)
                if item:
                    units += item.get_size().as_units()
                    items.append(item)
            push_read_throughput(tablename, base_capacity*units)
            ret[tablename][u'Items'] = items
            ret[tablename][u'ConsumedCapacityUnits'] = base_capacity*units

        return ret

    def write_batch(self, batch):
        """
        Batch processor. Dispatches call to appropriate :py:class:`ddbmock.database.table.Table`
        methods. This is the only low_level API that directly pushes throughput usage.

        :param batch: raw DynamoDB request batch.

        :returns: dict compatible with DynamoDB API

        :raises: :py:exc:`ddbmock.errors.ValidationException` if a ``range_key`` was provided while table has none.
        :raises: :py:exc:`ddbmock.errors.ResourceNotFoundException` if a table does not exist.
        """
        ret = defaultdict(dict)

        for tablename, operations in batch.iteritems():
            table = self.get_table(tablename)
            units = ItemSize(0)
            for operation in operations:
                if u'PutRequest' in operation:
                    old, new = table.put(operation[u'PutRequest'][u'Item'], {})
                    units += max(old.get_size().as_units(), new.get_size().as_units())
                if u'DeleteRequest' in operation:
                    old = table.delete_item(operation[u'DeleteRequest'][u'Key'], {})
                    units += old.get_size().as_units()
            push_write_throughput(tablename, units)
            ret[tablename][u'ConsumedCapacityUnits'] = units

        return ret

dynamodb = DynamoDB()
"""Reference :py:class:`DynamoDB` instance. You should use it directly"""



# -*- coding: utf-8 -*-

from . import load_table

@load_table
def update_table(post, table):
    table.update_throughput(post[u'ProvisionedThroughput'][u'ReadCapacityUnits'],
                            post[u'ProvisionedThroughput'][u'WriteCapacityUnits'],
                           )

    desc = table.to_dict()

    return {
        "TableDescription": desc,
    }
